#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Collection of tests for :mod:`orion.core.io.experiment_branch_builder`."""

import copy
import os

import pytest
import yaml

import orion.core
from orion.core import evc
from orion.core.evc.conflicts import (
    AlgorithmConflict, ChangedDimensionConflict, CodeConflict, CommandLineConflict,
    detect_conflicts, ExperimentNameConflict, MissingDimensionConflict, NewDimensionConflict,
    ScriptConfigConflict)
from orion.core.io.experiment_branch_builder import ExperimentBranchBuilder
import orion.core.utils.backward as backward


def filter_true(c):
    """Filter solved conflicts"""
    return c.is_resolved is True


def filter_false(c):
    """Filter unsolved conflicts"""
    return not filter_true(c)


@pytest.fixture
def user_config():
    """Generate data dict for user's script's configuration file"""
    data = {
        'a': 'orion~uniform(-10,10)',
        'some_other': 'test',
        'b': 'orion~normal(0,1)',
        'argument': 'value'}
    return data


@pytest.fixture
def parent_config(user_config):
    """Create a configuration that will not hit the database."""
    user_script = 'abs_path/black_box.py'
    config = dict(
        _id='test',
        name='test',
        algorithms='random',
        version=1,
        metadata={'VCS': {"type": "git",
                          "is_dirty": False,
                          "HEAD_sha": "test",
                          "active_branch": None,
                          "diff_sha": "diff",
                          },
                  'user_script': user_script,
                  'user_args': [
                      user_script, '--nameless=option', '-x~uniform(0,1)', '-y~normal(0,1)',
                      '-z~uniform(0,10)', '--manual-resolution'],
                  'user': 'some_user_name'},
        refers={})

    config_file_path = "./parent_config.yaml"

    with open(config_file_path, 'w') as f:
        yaml.dump(user_config, f)

    config['metadata']['user_args'].append('--config=%s' % config_file_path)

    backward.populate_space(config)

    yield config
    os.remove(config_file_path)


@pytest.fixture
def child_config(parent_config, create_db_instance):
    """Create a child branching from the test experiment"""
    config = copy.deepcopy(parent_config)
    config['_id'] = 'test2'
    config['version'] = 2
    config['refers']['parent_id'] = parent_config['_id']
    return config


@pytest.fixture
def missing_config(child_config):
    """Create a child config with a missing dimension"""
    del(child_config['metadata']['user_args'][2])  # del -x
    del(child_config['metadata']['user_args'][2])  # del -y
    backward.populate_space(child_config)
    return child_config


@pytest.fixture
def new_config(child_config):
    """Create a child config with a new dimension"""
    child_config['metadata']['user_args'].append('-w_d~normal(0,1)')
    backward.populate_space(child_config)
    return child_config


@pytest.fixture
def changed_config(child_config):
    """Create a child config with a changed dimension"""
    second_element = child_config['metadata']['user_args'][3]
    second_element = second_element.replace('normal', 'uniform')
    child_config['metadata']['user_args'][3] = second_element
    backward.populate_space(child_config)
    return child_config


@pytest.fixture
def changed_algo_config(child_config):
    """Create a child config with a changed dimension"""
    child_config['algorithms'] = 'stupid-grid'
    return child_config


@pytest.fixture
def changed_code_config(child_config):
    """Create a child config with a changed dimension"""
    child_config['metadata']['VCS']['HEAD_sha'] = 'new_test'
    return child_config


@pytest.fixture
def same_userconfig_config(user_config, child_config):
    """Create a child config with a changed dimension"""
    config_file_path = "./same_config.yaml"
    with open(config_file_path, 'w') as f:
        yaml.dump(user_config, f)
    child_config['metadata']['user_args'][-1] = '--config=%s' % config_file_path
    backward.populate_space(child_config)
    yield child_config
    os.remove(config_file_path)


@pytest.fixture
def changed_userconfig_config(user_config, child_config):
    """Create a child config with a changed dimension"""
    config_file_path = './changed_config.yaml'
    user_config['b'] = 'orion~uniform(-20, 0, precision=None)'
    user_config['some_other'] = 'hello'
    with open(config_file_path, 'w') as f:
        yaml.dump(user_config, f)
    child_config['metadata']['user_args'][-1] = '--config=%s' % config_file_path
    backward.populate_space(child_config)
    yield child_config
    os.remove(config_file_path)


@pytest.fixture
def changed_cli_config(child_config):
    """Create a child config with a changed dimension"""
    child_config['metadata']['user_args'] += ['-u=0', '--another=test', 'positional']
    backward.populate_space(child_config)
    return child_config


@pytest.fixture
def list_arg_with_equals_cli_config(child_config):
    """Create a child config with an argument of the
    form --args=1 --args=2 --args=3
    """
    child_config['metadata']['user_args'] += ['--args=1',
                                              '--args=2', '--args=3']
    backward.populate_space(child_config)
    return child_config


@pytest.fixture
def cl_config(create_db_instance):
    """Create a child config with markers for commandline solving"""
    user_script = 'abs_path/black_box.py'
    config = dict(
        name='test',
        branch='test2',
        algorithms='random',
        metadata={'hash_commit': 'old',
                  'user_script': user_script,
                  'user_args': [
                      user_script, '--nameless=option', '-x~>w_d', '-w_d~+normal(0,1)',
                      '-y~+uniform(0,1)', '-z~-', '--omega~+normal(0,1)'],
                  'user': 'some_user_name'})
    backward.populate_space(config)
    return config


@pytest.fixture
def conflicts(new_dimension_conflict, changed_dimension_conflict, missing_dimension_conflict,
              algorithm_conflict, code_conflict, experiment_name_conflict, config_conflict,
              cli_conflict):
    """Create a container for conflicts with one of each types for testing purposes"""
    conflicts = evc.conflicts.Conflicts()
    conflicts.register(new_dimension_conflict)
    conflicts.register(changed_dimension_conflict)
    conflicts.register(missing_dimension_conflict)
    conflicts.register(algorithm_conflict)
    conflicts.register(code_conflict)
    conflicts.register(experiment_name_conflict)
    conflicts.register(config_conflict)
    conflicts.register(cli_conflict)
    return conflicts


class TestConflictDetection(object):
    """Test detection of conflicts between two configurations"""

    def test_no_conflicts(self, parent_config, child_config):
        """Test the case where the child is the same as the parent"""
        conflicts = detect_conflicts(parent_config, child_config)

        assert len(conflicts.get()) == 1
        assert isinstance(conflicts.get()[0], ExperimentNameConflict)

    def test_missing_dim_conflict(self, parent_config, missing_config):
        """Test if missing dimension is currently detected"""
        conflicts = detect_conflicts(parent_config, missing_config)

        assert len(conflicts.get()) == 3
        conflict = conflicts.get()[1]

        assert conflict.is_resolved is False
        assert conflict.dimension.name == '/x'
        assert isinstance(conflict, MissingDimensionConflict)

    def test_new_dim_conflict(self, parent_config, new_config):
        """Test if new dimension is currently detected"""
        conflicts = detect_conflicts(parent_config, new_config)

        assert len(conflicts.get()) == 2
        conflict = conflicts.get()[1]

        assert conflict.is_resolved is False
        assert conflict.dimension.name == '/w_d'
        assert isinstance(conflict, NewDimensionConflict)

    def test_changed_dim_conflict(self, parent_config, changed_config):
        """Test if changed dimension is currently detected"""
        conflicts = detect_conflicts(parent_config, changed_config)

        assert len(conflicts.get()) == 2
        conflict = conflicts.get()[0]

        assert conflict.is_resolved is False
        assert conflict.dimension.name == '/y'
        assert isinstance(conflict, ChangedDimensionConflict)

    def test_changed_dim_userconfig_conflict(self, parent_config, changed_userconfig_config):
        """Test if changed dimension from user's config is currently detected"""
        conflicts = detect_conflicts(parent_config, changed_userconfig_config)

        assert len(conflicts.get()) == 4
        conflict = conflicts.get()[0]

        assert conflict.is_resolved is False
        assert conflict.dimension.name == '/b'
        assert isinstance(conflict, ChangedDimensionConflict)

    def test_algo_conflict(self, parent_config, changed_algo_config):
        """Test if algorithm changes are currently detected"""
        conflicts = detect_conflicts(parent_config, changed_algo_config)

        assert len(conflicts.get()) == 2
        conflict = conflicts.get()[0]

        assert conflict.is_resolved is False
        assert conflict.old_config['algorithms'] == 'random'
        assert conflict.new_config['algorithms'] == 'stupid-grid'
        assert isinstance(conflict, AlgorithmConflict)

    def test_code_conflict(self, parent_config, changed_code_config):
        """Test if code commit hash change is currently detected"""
        conflicts = detect_conflicts(parent_config, changed_code_config)

        assert len(conflicts.get()) == 2
        conflict = conflicts.get()[0]

        assert conflict.is_resolved is False
        assert conflict.old_config['metadata']['VCS']['HEAD_sha'] == 'test'
        assert conflict.new_config['metadata']['VCS']['HEAD_sha'] == 'new_test'
        assert isinstance(conflict, CodeConflict)

    def test_ignore_code_conflict(self, parent_config, changed_code_config):
        """Test if ignored code commit hash change is detected as a conflict"""
        conflicts = detect_conflicts(parent_config, changed_code_config,
                                     {'ignore_code_changes': True})

        assert len(conflicts.get()) == 1

    def test_config_new_name_no_conflict(self, parent_config, same_userconfig_config):
        """Test if same configuration file with a different name is not detected as a conflict"""
        conflicts = detect_conflicts(parent_config, same_userconfig_config)

        assert parent_config['metadata']['user_args'][-1].startswith('--config=')
        assert same_userconfig_config['metadata']['user_args'][-1].startswith('--config=')
        assert (parent_config['metadata']['user_args'][-1] !=
                same_userconfig_config['metadata']['user_args'][-1])

        assert len(conflicts.get()) == 2
        assert not conflicts.get([ExperimentNameConflict])[0].is_resolved

    def test_config_non_dim_conflict(self, parent_config, changed_userconfig_config):
        """Test if changed configuration file is detected as a conflict"""
        conflicts = detect_conflicts(parent_config, changed_userconfig_config)

        assert len(conflicts.get()) == 4
        assert not conflicts.get([ChangedDimensionConflict])[0].is_resolved
        assert not conflicts.get([ExperimentNameConflict])[0].is_resolved
        assert not conflicts.get([ScriptConfigConflict])[0].is_resolved

    def test_cli_conflict(self, parent_config, changed_cli_config):
        """Test if changed command line call is detected as a conflict"""
        conflicts = detect_conflicts(parent_config, changed_cli_config)

        assert len(conflicts.get()) == 2

        assert not conflicts.get([ExperimentNameConflict])[0].is_resolved
        assert not conflicts.get([CommandLineConflict])[0].is_resolved

    def test_cli_ignored_conflict(self, parent_config, changed_cli_config):
        """Test if ignored changed command line call is detected as a conflict"""
        changed_cli_config['metadata']['user_args'].pop()
        conflicts = detect_conflicts(parent_config, changed_cli_config,
                                     {'non_monitored_arguments': ['u', 'another']})

        assert len(conflicts.get()) == 1

        assert not conflicts.get([ExperimentNameConflict])[0].is_resolved


class TestResolutions(object):
    """Test resolution of conflicts"""

    def test_add_single_hit(self, parent_config, new_config):
        """Test if adding a dimension only touches the correct status"""
        del new_config['metadata']['user_args'][2]
        backward.populate_space(new_config)
        conflicts = detect_conflicts(parent_config, new_config)
        branch_builder = ExperimentBranchBuilder(conflicts, manual_resolution=True)
        branch_builder.add_dimension('w_d')

        assert len(conflicts.get()) == 3
        assert conflicts.get([ExperimentNameConflict])[0].is_resolved
        assert conflicts.get([NewDimensionConflict])[0].is_resolved
        assert not conflicts.get([MissingDimensionConflict])[0].is_resolved

    def test_add_new(self, parent_config, new_config):
        """Test if adding a new dimension solves the conflict"""
        conflicts = detect_conflicts(parent_config, new_config)
        branch_builder = ExperimentBranchBuilder(conflicts, manual_resolution=True)
        branch_builder.add_dimension('w_d')

        assert len(conflicts.get()) == 2
        assert len(conflicts.get_resolved()) == 2

        conflict = conflicts.get_resolved()[1]

        assert conflict.is_resolved
        assert isinstance(conflict.resolution, conflict.AddDimensionResolution)

    def test_add_changed(self, parent_config, changed_config):
        """Test if adding a changed dimension solves the conflict"""
        conflicts = detect_conflicts(parent_config, changed_config)
        branch_builder = ExperimentBranchBuilder(conflicts, manual_resolution=True)
        branch_builder.add_dimension('y')

        assert len(conflicts.get()) == 2
        assert len(conflicts.get_resolved()) == 2

        conflict = conflicts.get_resolved()[0]

        assert conflict.is_resolved
        assert isinstance(conflict.resolution, conflict.ChangeDimensionResolution)

    def test_remove_missing(self, parent_config, missing_config):
        """Test if removing a missing dimension solves the conflict"""
        conflicts = detect_conflicts(parent_config, missing_config)
        branch_builder = ExperimentBranchBuilder(conflicts, manual_resolution=True)
        branch_builder.remove_dimension('x')

        assert len(conflicts.get()) == 3
        assert len(conflicts.get_resolved()) == 2

        conflict = conflicts.get_resolved()[1]

        assert conflict.is_resolved
        assert isinstance(conflict.resolution, conflict.RemoveDimensionResolution)

    def test_rename_missing(self, parent_config, missing_config):
        """Test if renaming a dimension to another solves both conflicts"""
        missing_config['metadata']['user_args'].append('-w_d~uniform(0,1)')
        backward.populate_space(missing_config)
        conflicts = detect_conflicts(parent_config, missing_config)
        branch_builder = ExperimentBranchBuilder(conflicts, manual_resolution=True)
        branch_builder.rename_dimension('x', 'w_d')

        assert len(conflicts.get()) == 4

        assert conflicts.get([ExperimentNameConflict])[0].is_resolved
        assert conflicts.get([NewDimensionConflict])[0].is_resolved
        assert conflicts.get([MissingDimensionConflict])[0].is_resolved
        assert not conflicts.get([MissingDimensionConflict])[1].is_resolved

        resolved_conflicts = conflicts.get_resolved()
        assert len(resolved_conflicts) == 3
        assert resolved_conflicts[1].resolution is resolved_conflicts[2].resolution
        assert isinstance(resolved_conflicts[1].resolution,
                          resolved_conflicts[1].RenameDimensionResolution)
        assert resolved_conflicts[1].resolution.conflict.dimension.name == '/x'
        assert resolved_conflicts[1].resolution.new_dimension_conflict.dimension.name == '/w_d'

    def test_rename_missing_changed(self, parent_config, missing_config):
        """Test if renaming a dimension to another with different prior solves both conflicts but
        creates a new one which is not solved
        """
        missing_config['metadata']['user_args'].append('-w_d~normal(0,1)')
        backward.populate_space(missing_config)
        conflicts = detect_conflicts(parent_config, missing_config)
        branch_builder = ExperimentBranchBuilder(conflicts, manual_resolution=True)

        assert len(conflicts.get()) == 4

        branch_builder.rename_dimension('x', 'w_d')

        assert len(conflicts.get()) == 5

        assert conflicts.get([ExperimentNameConflict])[0].is_resolved
        assert conflicts.get([NewDimensionConflict])[0].is_resolved
        assert conflicts.get([MissingDimensionConflict])[0].is_resolved
        assert not conflicts.get([MissingDimensionConflict])[1].is_resolved
        assert not conflicts.get([ChangedDimensionConflict])[0].is_resolved

        resolved_conflicts = conflicts.get_resolved()
        assert len(resolved_conflicts) == 3
        assert resolved_conflicts[1].resolution is resolved_conflicts[2].resolution
        assert isinstance(resolved_conflicts[1].resolution,
                          resolved_conflicts[1].RenameDimensionResolution)
        assert resolved_conflicts[1].resolution.conflict.dimension.name == '/x'
        assert resolved_conflicts[1].resolution.new_dimension_conflict.dimension.name == '/w_d'

    def test_reset_dimension(self, parent_config, new_config):
        """Test if resetting a dimension unsolves the conflict"""
        conflicts = detect_conflicts(parent_config, new_config)
        branch_builder = ExperimentBranchBuilder(conflicts, manual_resolution=True)

        branch_builder.add_dimension('w_d')
        assert len(conflicts.get_resolved()) == 2

        with pytest.raises(ValueError) as exc:
            branch_builder.reset('w_d~+')
        assert "'w_d~+' is not in list" in str(exc.value)
        assert len(conflicts.get_resolved()) == 2

        branch_builder.reset('w_d~+norm(0, 1)')

        assert len(conflicts.get()) == 2

        conflict = conflicts.get(dimension_name='w_d')[0]

        assert not conflict.is_resolved
        assert isinstance(conflict, NewDimensionConflict)
        assert len(conflicts.get_resolved()) == 1

    def test_name_experiment(self, bad_exp_parent_config, bad_exp_child_config, create_db_instance):
        """Test if having the same experiment name does not create a conflict."""
        backward.populate_space(bad_exp_parent_config)
        backward.populate_space(bad_exp_child_config)
        create_db_instance.write('experiments', bad_exp_parent_config)
        create_db_instance.write('experiments', bad_exp_child_config)
        conflicts = detect_conflicts(bad_exp_parent_config, bad_exp_parent_config)
        branch_builder = ExperimentBranchBuilder(conflicts, manual_resolution=True)

        assert len(conflicts.get()) == 1
        assert len(conflicts.get_resolved()) == 0

        conflict = conflicts.get([ExperimentNameConflict])[0]

        assert conflict.new_config['name'] == 'test'
        assert not conflict.is_resolved
        branch_builder.change_experiment_name('test2')
        assert len(conflicts.get_resolved()) == 1
        assert conflict.new_config['name'] == 'test2'
        assert conflict.is_resolved

    def test_bad_name_experiment(self, parent_config, child_config, monkeypatch):
        """Test if changing the experiment names does not work for invalid name and revert
        to old one
        """
        def _is_unique(self, *args, **kwargs):
            return False

        def _versions(self, *args, **kwargs):
            return True

        monkeypatch.setattr(ExperimentNameConflict.ExperimentNameResolution, "_name_is_unique",
                            _is_unique)
        monkeypatch.setattr(ExperimentNameConflict.ExperimentNameResolution,
                            "_check_for_greater_versions", _versions)

        conflicts = detect_conflicts(parent_config, child_config)
        branch_builder = ExperimentBranchBuilder(conflicts, manual_resolution=True)

        assert len(conflicts.get()) == 1
        assert len(conflicts.get_resolved()) == 0

        conflict = conflicts.get([ExperimentNameConflict])[0]

        assert not conflict.is_resolved
        branch_builder.change_experiment_name('test2')
        assert len(conflicts.get_resolved()) == 0
        assert conflict.new_config['name'] == 'test'
        assert not conflict.is_resolved

    def test_algo_change(self, parent_config, changed_algo_config):
        """Test if setting the algorithm conflict solves it"""
        conflicts = detect_conflicts(parent_config, changed_algo_config)
        branch_builder = ExperimentBranchBuilder(conflicts, manual_resolution=True)

        assert len(conflicts.get()) == 2
        assert len(conflicts.get_resolved()) == 1

        branch_builder.set_algo()

        assert len(conflicts.get()) == 2
        assert len(conflicts.get_resolved()) == 2

        conflict = conflicts.get_resolved()[0]

        assert conflict.is_resolved
        assert isinstance(conflict.resolution, conflict.AlgorithmResolution)

    def test_code_change(self, parent_config, changed_code_config):
        """Test if giving a proper change-type solves the code conflict"""
        conflicts = detect_conflicts(parent_config, changed_code_config)
        branch_builder = ExperimentBranchBuilder(conflicts, manual_resolution=True)

        assert len(conflicts.get()) == 2
        assert len(conflicts.get_resolved()) == 1

        branch_builder.set_code_change_type(evc.adapters.CodeChange.types[0])

        assert len(conflicts.get()) == 2
        assert len(conflicts.get_resolved()) == 2

        conflict = conflicts.get_resolved()[0]
        assert conflict.is_resolved
        assert isinstance(conflict, CodeConflict)

    def test_bad_code_change(self, capsys, parent_config, changed_code_config):
        """Test if giving an invalid change-type prints error message and do nothing"""
        conflicts = detect_conflicts(parent_config, changed_code_config)
        branch_builder = ExperimentBranchBuilder(conflicts, manual_resolution=True)
        capsys.readouterr()
        branch_builder.set_code_change_type('bad-type')
        out, err = capsys.readouterr()
        assert 'Invalid code change type' in out.split("\n")[-3]

        assert len(conflicts.get()) == 2
        assert len(conflicts.get_resolved()) == 1

    def test_config_change(self, parent_config, changed_userconfig_config):
        """Test if giving a proper change-type solves the user script config conflict"""
        conflicts = detect_conflicts(parent_config, changed_userconfig_config)
        branch_builder = ExperimentBranchBuilder(conflicts, manual_resolution=True)

        assert len(conflicts.get()) == 4
        assert len(conflicts.get_resolved()) == 1

        branch_builder.set_script_config_change_type(evc.adapters.ScriptConfigChange.types[0])

        assert len(conflicts.get()) == 4
        assert len(conflicts.get_resolved()) == 2

        conflict = conflicts.get_resolved()[1]
        assert conflict.is_resolved
        assert isinstance(conflict, ScriptConfigConflict)

    def test_bad_config_change(self, capsys, parent_config, changed_userconfig_config):
        """Test if giving an invalid change-type prints error message and do nothing"""
        conflicts = detect_conflicts(parent_config, changed_userconfig_config)
        branch_builder = ExperimentBranchBuilder(conflicts, manual_resolution=True)
        capsys.readouterr()
        branch_builder.set_script_config_change_type('bad-type')
        out, err = capsys.readouterr()
        assert 'Invalid script\'s config change type' in out.split("\n")[-3]

        assert len(conflicts.get()) == 4
        assert len(conflicts.get_resolved()) == 1

    def test_cli_change(self, parent_config, changed_cli_config):
        """Test if giving a proper change-type solves the command line conflict"""
        conflicts = detect_conflicts(parent_config, changed_cli_config)
        branch_builder = ExperimentBranchBuilder(conflicts, manual_resolution=True)

        assert len(conflicts.get()) == 2
        assert len(conflicts.get_resolved()) == 1

        branch_builder.set_cli_change_type(evc.adapters.CommandLineChange.types[0])

        assert len(conflicts.get()) == 2
        assert len(conflicts.get_resolved()) == 2

        conflict = conflicts.get_resolved()[0]
        assert conflict.is_resolved
        assert isinstance(conflict, CommandLineConflict)

    def test_bad_cli_change(self, capsys, parent_config, changed_cli_config):
        """Test if giving an invalid change-type prints error message and do nothing"""
        conflicts = detect_conflicts(parent_config, changed_cli_config)
        branch_builder = ExperimentBranchBuilder(conflicts, manual_resolution=True)
        capsys.readouterr()
        branch_builder.set_cli_change_type('bad-type')
        out, err = capsys.readouterr()
        assert 'Invalid cli change type' in out.split("\n")[-3]

        assert len(conflicts.get()) == 2
        assert len(conflicts.get_resolved()) == 1

    def test_solve_all_automatically(self, conflicts):
        """Test if all conflicts all automatically resolve by the ExperimentBranchBuilder."""
        ExperimentBranchBuilder(conflicts)

        assert len(conflicts.get_resolved()) == 8


class TestResolutionsWithMarkers(object):
    """Test resolution of conflicts with markers"""

    def test_add_new(self, parent_config, new_config):
        """Test if new dimension conflict is automatically resolved"""
        new_config['metadata']['user_args'][-1] = '-w_d~+normal(0,1)'
        conflicts = detect_conflicts(parent_config, new_config)
        ExperimentBranchBuilder(conflicts, manual_resolution=True)

        assert len(conflicts.get()) == 2
        assert len(conflicts.get_resolved()) == 2

        conflict = conflicts.get_resolved()[1]

        assert conflict.is_resolved
        assert isinstance(conflict.resolution, conflict.AddDimensionResolution)

    def test_add_new_default(self, parent_config, new_config):
        """Test if new dimension conflict is automatically resolved"""
        new_config['metadata']['user_args'][-1] = '-w_d~+normal(0,1,default_value=0)'
        backward.populate_space(new_config)
        conflicts = detect_conflicts(parent_config, new_config)
        ExperimentBranchBuilder(conflicts, manual_resolution=True)

        assert len(conflicts.get()) == 2
        assert len(conflicts.get_resolved()) == 2

        conflict = conflicts.get_resolved()[1]

        assert conflict.is_resolved
        assert isinstance(conflict.resolution, conflict.AddDimensionResolution)
        assert conflict.resolution.default_value == 0.0

    def test_add_bad_default(self, parent_config, new_config):
        """Test if new dimension conflict raises an error if marked with invalid default value"""
        new_config['metadata']['user_args'][-1] = '-w_d~+normal(0,1,default_value=\'a\')'
        backward.populate_space(new_config)
        with pytest.raises(TypeError) as exc:
            detect_conflicts(parent_config, new_config)
        assert "Parameter \'/w_d\': Incorrect arguments." in str(exc.value)

    def test_add_changed(self, parent_config, changed_config):
        """Test if changed dimension conflict is automatically resolved"""
        changed_config['metadata']['user_args'][3] = (
            changed_config['metadata']['user_args'][3].replace("~", "~+"))
        conflicts = detect_conflicts(parent_config, changed_config)
        ExperimentBranchBuilder(conflicts, manual_resolution=True)

        assert len(conflicts.get()) == 2
        assert len(conflicts.get_resolved()) == 2

        conflict = conflicts.get_resolved()[0]

        assert conflict.is_resolved
        assert isinstance(conflict.resolution, conflict.ChangeDimensionResolution)

    def test_remove_missing(self, parent_config, child_config):
        """Test if missing dimension conflict is automatically resolved"""
        child_config['metadata']['user_args'][2] = '-x~-'
        backward.populate_space(child_config)
        conflicts = detect_conflicts(parent_config, child_config)
        ExperimentBranchBuilder(conflicts, manual_resolution=True)

        assert len(conflicts.get()) == 2
        assert len(conflicts.get_resolved()) == 2

        conflict = conflicts.get_resolved()[1]

        assert conflict.is_resolved
        assert isinstance(conflict.resolution, conflict.RemoveDimensionResolution)

    def test_remove_missing_default(self, parent_config, child_config):
        """Test if missing dimension conflict is automatically resolved"""
        child_config['metadata']['user_args'][2] = '-x~-0.5'
        backward.populate_space(child_config)
        conflicts = detect_conflicts(parent_config, child_config)
        ExperimentBranchBuilder(conflicts, manual_resolution=True)

        assert len(conflicts.get()) == 2
        assert len(conflicts.get_resolved()) == 2

        conflict = conflicts.get_resolved()[1]

        assert conflict.is_resolved
        assert isinstance(conflict.resolution, conflict.RemoveDimensionResolution)
        assert conflict.resolution.default_value == 0.5

    def test_remove_missing_bad_default(self, parent_config, child_config):
        """Test if missing dimension conflict raises an error if marked with invalid default"""
        child_config['metadata']['user_args'][2] = '-x~--100'
        backward.populate_space(child_config)
        conflicts = detect_conflicts(parent_config, child_config)
        ExperimentBranchBuilder(conflicts, manual_resolution=True)

        assert len(conflicts.get()) == 2
        assert len(conflicts.get_resolved()) == 1

        conflict = conflicts.get()[1]

        assert not conflict.is_resolved
        assert isinstance(conflict, MissingDimensionConflict)

    def test_rename_missing(self, parent_config, child_config):
        """Test if renaming is automatically applied with both conflicts resolved"""
        child_config['metadata']['user_args'].append('-w_a~uniform(0,1)')
        child_config['metadata']['user_args'].append('-w_b~normal(0,1)')
        child_config['metadata']['user_args'][2] = '-x~>w_a'
        backward.populate_space(child_config)
        conflicts = detect_conflicts(parent_config, child_config)
        ExperimentBranchBuilder(conflicts, manual_resolution=True)

        assert len(conflicts.get()) == 4

        assert conflicts.get([ExperimentNameConflict])[0].is_resolved
        assert conflicts.get(dimension_name='x')[0].is_resolved
        assert conflicts.get(dimension_name='w_a')[0].is_resolved
        assert not conflicts.get(dimension_name='w_b')[0].is_resolved

        resolved_conflicts = conflicts.get_resolved()
        assert len(resolved_conflicts) == 3
        assert resolved_conflicts[1].resolution is resolved_conflicts[2].resolution
        assert isinstance(resolved_conflicts[1].resolution,
                          resolved_conflicts[1].RenameDimensionResolution)
        assert resolved_conflicts[1].resolution.conflict.dimension.name == '/x'
        assert resolved_conflicts[1].resolution.new_dimension_conflict.dimension.name == '/w_a'

    def test_rename_invalid(self, parent_config, child_config):
        """Test if renaming to invalid dimension raises an error"""
        child_config['metadata']['user_args'].append('-w_a~uniform(0,1)')
        child_config['metadata']['user_args'].append('-w_b~uniform(0,1)')
        child_config['metadata']['user_args'][2] = '-x~>w_c'
        backward.populate_space(child_config)
        conflicts = detect_conflicts(parent_config, child_config)
        with pytest.raises(ValueError) as exc:
            ExperimentBranchBuilder(conflicts, manual_resolution=True)
        assert "Dimension name 'w_c' not found in conflicts" in str(exc.value)

    def test_rename_missing_changed(self, parent_config, child_config):
        """Test if renaming is automatically applied with both conflicts resolved,
        but not the new one because of prior change
        """
        child_config['metadata']['user_args'].append('-w_a~uniform(0,1)')
        child_config['metadata']['user_args'].append('-w_b~normal(0,1)')
        child_config['metadata']['user_args'][2] = '-x~>w_b'
        backward.populate_space(child_config)
        conflicts = detect_conflicts(parent_config, child_config)
        ExperimentBranchBuilder(conflicts, manual_resolution=True)

        assert len(conflicts.get()) == 5

        assert conflicts.get([ExperimentNameConflict])[0].is_resolved
        assert conflicts.get(dimension_name='x')[0].is_resolved
        assert conflicts.get([NewDimensionConflict], dimension_name='w_b')[0].is_resolved
        assert not conflicts.get([ChangedDimensionConflict], dimension_name='w_b')[0].is_resolved
        assert not conflicts.get(dimension_name='w_a')[0].is_resolved

        resolved_conflicts = conflicts.get_resolved()
        assert len(resolved_conflicts) == 3
        assert resolved_conflicts[1].resolution is resolved_conflicts[2].resolution
        assert isinstance(resolved_conflicts[1].resolution,
                          resolved_conflicts[1].RenameDimensionResolution)
        assert resolved_conflicts[1].resolution.conflict.dimension.name == '/x'
        assert resolved_conflicts[1].resolution.new_dimension_conflict.dimension.name == '/w_b'

    def test_rename_missing_changed_marked(self, parent_config, child_config):
        """Test if renaming is automatically applied with all conflicts resolved including
        the new one caused by prior change
        """
        child_config['metadata']['user_args'].append('-w_a~uniform(0,1)')
        child_config['metadata']['user_args'].append('-w_b~+normal(0,1)')
        child_config['metadata']['user_args'][2] = '-x~>w_b'
        backward.populate_space(child_config)
        conflicts = detect_conflicts(parent_config, child_config)
        ExperimentBranchBuilder(conflicts, manual_resolution=True)

        assert len(conflicts.get()) == 5

        assert conflicts.get([ExperimentNameConflict])[0].is_resolved
        assert conflicts.get(dimension_name='x')[0].is_resolved
        assert conflicts.get([NewDimensionConflict], dimension_name='w_b')[0].is_resolved
        assert conflicts.get([ChangedDimensionConflict], dimension_name='w_b')[0].is_resolved
        assert not conflicts.get(dimension_name='w_a')[0].is_resolved

        resolved_conflicts = conflicts.get_resolved()
        assert len(resolved_conflicts) == 4
        assert resolved_conflicts[1].resolution is resolved_conflicts[2].resolution
        assert isinstance(resolved_conflicts[1].resolution,
                          resolved_conflicts[1].RenameDimensionResolution)
        assert resolved_conflicts[1].resolution.conflict.dimension.name == '/x'
        assert resolved_conflicts[1].resolution.new_dimension_conflict.dimension.name == '/w_b'

    def test_name_experiment_version_update(self, parent_config, child_config, create_db_instance):
        """Test if experiment name conflict is automatically resolved with version update"""
        old_name = 'test'
        new_version = 2
        create_db_instance.write('experiments', parent_config)
        child_config['version'] = 1
        conflicts = detect_conflicts(parent_config, child_config)
        ExperimentBranchBuilder(conflicts)

        assert len(conflicts.get()) == 1
        assert len(conflicts.get_resolved()) == 1

        conflict = conflicts.get()[0]

        assert conflict.resolution.new_name == old_name
        assert conflict.resolution.new_version == 2
        assert conflict.new_config['name'] == old_name
        assert conflict.new_config['version'] == new_version
        assert conflict.is_resolved

    def test_name_experiment_name_change(self, parent_config, child_config, create_db_instance):
        """Test if experiment name conflict is automatically resolved when new name provided"""
        new_name = 'test2'
        create_db_instance.write('experiments', parent_config)
        create_db_instance.write('experiments', child_config)
        child_config2 = copy.deepcopy(child_config)
        child_config2['version'] = 1
        conflicts = detect_conflicts(parent_config, child_config2)
        ExperimentBranchBuilder(conflicts, branch_to=new_name)

        assert len(conflicts.get()) == 1
        assert len(conflicts.get_resolved()) == 1

        conflict = conflicts.get()[0]

        assert conflict.resolution.new_name == new_name
        assert conflict.resolution.new_version == 1
        assert conflict.new_config['name'] == new_name
        assert conflict.new_config['version'] == 1
        assert conflict.is_resolved

    def test_bad_name_experiment(self, parent_config, child_config, monkeypatch):
        """Test if experiment name conflict is not resolved when invalid name is marked"""
        def _is_unique(self, *args, **kwargs):
            return False

        monkeypatch.setattr(ExperimentNameConflict.ExperimentNameResolution, "_name_is_unique",
                            _is_unique)

        conflicts = detect_conflicts(parent_config, child_config)
        ExperimentBranchBuilder(conflicts, branch_to='test2')

        assert len(conflicts.get()) == 1
        assert len(conflicts.get_resolved()) == 0

    def test_code_change(self, parent_config, changed_code_config):
        """Test if code conflict is resolved automatically"""
        change_type = evc.adapters.CodeChange.types[0]
        changed_code_config['code_change_type'] = change_type
        conflicts = detect_conflicts(parent_config, changed_code_config)
        ExperimentBranchBuilder(conflicts, manual_resolution=True)

        assert len(conflicts.get()) == 2
        assert len(conflicts.get_resolved()) == 2

        conflict = conflicts.get_resolved()[0]

        assert conflict.is_resolved
        assert isinstance(conflict.resolution, conflict.CodeResolution)
        assert conflict.resolution.type == change_type

    def test_algo_change(self, parent_config, changed_algo_config):
        """Test if algorithm conflict is resolved automatically"""
        changed_algo_config['algorithm_change'] = True
        conflicts = detect_conflicts(parent_config, changed_algo_config)
        ExperimentBranchBuilder(conflicts, manual_resolution=True)

        assert len(conflicts.get()) == 2
        assert len(conflicts.get_resolved()) == 2

        conflict = conflicts.get_resolved()[0]

        assert conflict.is_resolved
        assert isinstance(conflict.resolution, conflict.AlgorithmResolution)

    def test_config_change(self, parent_config, changed_userconfig_config):
        """Test if user's script's config conflict is resolved automatically"""
        change_type = evc.adapters.ScriptConfigChange.types[0]
        changed_userconfig_config['config_change_type'] = change_type
        conflicts = detect_conflicts(parent_config, changed_userconfig_config)
        ExperimentBranchBuilder(conflicts, manual_resolution=True)

        assert len(conflicts.get()) == 4
        assert len(conflicts.get_resolved()) == 2

        conflict = conflicts.get_resolved()[1]

        assert conflict.is_resolved
        assert isinstance(conflict.resolution, conflict.ScriptConfigResolution)
        assert conflict.resolution.type == change_type

    def test_cli_change(self, parent_config, changed_cli_config):
        """Test if command line conflict is resolved automatically"""
        change_type = evc.adapters.CommandLineChange.types[0]
        changed_cli_config['cli_change_type'] = change_type
        conflicts = detect_conflicts(parent_config, changed_cli_config)
        ExperimentBranchBuilder(conflicts, manual_resolution=True)

        assert len(conflicts.get()) == 2
        assert len(conflicts.get_resolved()) == 2

        conflict = conflicts.get_resolved()[0]

        assert conflict.is_resolved
        assert isinstance(conflict.resolution, conflict.CommandLineResolution)
        assert conflict.resolution.type == change_type


class TestAdapters(object):
    """Test creation of adapters"""

    def test_adapter_add_new(self, parent_config, cl_config):
        """Test if a DimensionAddition is created when solving a new conflict"""
        cl_config['metadata']['user_args'] = ['-w_d~+normal(0,1)']

        conflicts = detect_conflicts(parent_config, cl_config)
        branch_builder = ExperimentBranchBuilder(conflicts, manual_resolution=True)

        adapters = branch_builder.create_adapters().adapters

        assert len(conflicts.get_resolved()) == 2
        assert len(adapters) == 1
        assert isinstance(adapters[0], evc.adapters.DimensionAddition)

    def test_adapter_add_changed(self, parent_config, cl_config):
        """Test if a DimensionPriorChange is created when solving a new conflict"""
        cl_config['metadata']['user_args'] = ['-y~+uniform(0,1)']

        conflicts = detect_conflicts(parent_config, cl_config)
        branch_builder = ExperimentBranchBuilder(conflicts, manual_resolution=True)

        adapters = branch_builder.create_adapters().adapters

        assert len(conflicts.get_resolved()) == 2
        assert len(adapters) == 1
        assert isinstance(adapters[0], evc.adapters.DimensionPriorChange)

    def test_adapter_remove_missing(self, parent_config, cl_config):
        """Test if a DimensionDeletion is created when solving a new conflict"""
        cl_config['metadata']['user_args'] = ['-z~-']

        conflicts = detect_conflicts(parent_config, cl_config)
        branch_builder = ExperimentBranchBuilder(conflicts, manual_resolution=True)

        adapters = branch_builder.create_adapters().adapters

        assert len(conflicts.get_resolved()) == 2
        assert len(adapters) == 1
        assert isinstance(adapters[0], evc.adapters.DimensionDeletion)

    def test_adapter_rename_missing(self, parent_config, cl_config):
        """Test if a DimensionRenaming is created when solving a new conflict"""
        cl_config['metadata']['user_args'] = ['-x~>w_d', '-w_d~+uniform(0,1)']
        backward.populate_space(cl_config)

        conflicts = detect_conflicts(parent_config, cl_config)
        branch_builder = ExperimentBranchBuilder(conflicts, manual_resolution=True)

        adapters = branch_builder.create_adapters().adapters

        assert len(conflicts.get_resolved()) == 3
        assert len(adapters) == 1
        assert isinstance(adapters[0], evc.adapters.DimensionRenaming)

    def test_adapter_rename_different_prior(self, parent_config, cl_config):
        """Test if a DimensionRenaming is created when solving a new conflict"""
        cl_config['metadata']['user_args'] = ['-x~>w_d', '-w_d~+normal(0,1)']

        conflicts = detect_conflicts(parent_config, cl_config)
        branch_builder = ExperimentBranchBuilder(conflicts, manual_resolution=True)

        adapters = branch_builder.create_adapters().adapters

        assert len(conflicts.get_resolved()) == 4
        assert len(adapters) == 2
        assert isinstance(adapters[0], evc.adapters.DimensionRenaming)
        assert isinstance(adapters[1], evc.adapters.DimensionPriorChange)


class TestResolutionsConfig(object):
    """Test auto-resolution with specific types from orion.core.config.evc"""

    def test_cli_change(self, parent_config, changed_cli_config):
        """Test if giving a proper change-type solves the command line conflict"""
        conflicts = detect_conflicts(parent_config, changed_cli_config)
        orion.core.config.evc.cli_change_type = 'noeffect'
        ExperimentBranchBuilder(conflicts)

        assert len(conflicts.get()) == 2
        assert len(conflicts.get_resolved()) == 2

        conflict = conflicts.get_resolved()[0]
        assert conflict.is_resolved
        assert isinstance(conflict, CommandLineConflict)
        assert conflict.resolution.type == 'noeffect'
        orion.core.config.evc.cli_change_type = 'break'

    def test_bad_cli_change(self, capsys, parent_config, changed_cli_config):
        """Test if giving an invalid change-type fails the the resolution"""
        conflicts = detect_conflicts(parent_config, changed_cli_config)
        orion.core.config.evc.cli_change_type = 'bad-type'
        ExperimentBranchBuilder(conflicts)

        assert len(conflicts.get()) == 2
        assert len(conflicts.get_resolved()) == 1
        orion.core.config.evc.cli_change_type = 'break'

    def test_code_change(self, parent_config, changed_code_config):
        """Test if giving a proper change-type solves the code conflict"""
        conflicts = detect_conflicts(parent_config, changed_code_config)
        orion.core.config.evc.code_change_type = 'noeffect'
        ExperimentBranchBuilder(conflicts)

        assert len(conflicts.get()) == 2
        assert len(conflicts.get_resolved()) == 2

        conflict = conflicts.get_resolved()[0]
        assert conflict.is_resolved
        assert isinstance(conflict, CodeConflict)
        assert conflict.resolution.type == 'noeffect'
        orion.core.config.evc.code_change_type = 'break'

    def test_bad_code_change(self, capsys, parent_config, changed_code_config):
        """Test if giving an invalid change-type prints error message and do nothing"""
        conflicts = detect_conflicts(parent_config, changed_code_config)
        orion.core.config.evc.code_change_type = 'bad-type'
        ExperimentBranchBuilder(conflicts)

        assert len(conflicts.get()) == 2
        assert len(conflicts.get_resolved()) == 1
        orion.core.config.evc.code_change_type = 'break'

    def test_config_change(self, parent_config, changed_userconfig_config):
        """Test if giving a proper change-type solves the user script config conflict"""
        conflicts = detect_conflicts(parent_config, changed_userconfig_config)
        orion.core.config.evc.config_change_type = 'noeffect'
        ExperimentBranchBuilder(conflicts)

        assert len(conflicts.get()) == 4
        assert len(conflicts.get_resolved()) == 4

        conflict = conflicts.get_resolved()[3]
        assert conflict.is_resolved
        assert isinstance(conflict, ScriptConfigConflict)
        assert conflict.resolution.type == 'noeffect'

    def test_bad_config_change(self, capsys, parent_config, changed_userconfig_config):
        """Test if giving an invalid change-type prints error message and do nothing"""
        conflicts = detect_conflicts(parent_config, changed_userconfig_config)
        orion.core.config.evc.config_change_type = 'bad-type'
        ExperimentBranchBuilder(conflicts)

        assert len(conflicts.get()) == 4
        assert len(conflicts.get_resolved()) == 3



"""
:mod:`orion.core.plotting.backend_plotly` -- Plotly backend for plotting methods
================================================================================

.. module:: backend_plotly
   :platform: Unix
   :synopsis: Plotly backend for plotting methods
"""
import pandas as pd
import plotly.graph_objects as go

import orion.analysis.regret


def regret(experiment, order_by, verbose_hover, **kwargs):
    """Plotly implementation of `orion.plotting.regret`"""
    def build_frame(trials):
        """Builds the dataframe for the plot"""
        data = [(trial.id, trial.status,
                trial.submit_time, trial.start_time, trial.end_time,
                _format_hyperparameters(trial.params), trial.objective.value) for trial in trials]

        df = pd.DataFrame(data, columns=['id', 'status', ORDER_KEYS[0],
                                         ORDER_KEYS[1], ORDER_KEYS[2], 'params', 'objective'])

        df = df.sort_values(order_by)
        df = orion.analysis.regret(df)
        return df

    ORDER_KEYS = ['suggested', 'reserved', 'completed']

    if not experiment:
        raise ValueError("Parameter 'experiment' is None")

    if order_by not in ORDER_KEYS:
        raise ValueError(f"Parameter 'order_by' is not one of {ORDER_KEYS}")

    trials = list(filter(lambda trial: trial.status == 'completed', experiment.fetch_trials()))
    df = build_frame(trials)

    fig = go.Figure()

    fig.add_scatter(y=df['objective'],
                    mode='markers',
                    name='trials',
                    customdata=list(zip(df['id'], df[order_by], df['params'])),
                    hovertemplate=_template_trials(verbose_hover))
    fig.add_scatter(y=df['best'],
                    mode='lines',
                    name='best-to-date',
                    customdata=list(zip(df['best_id'], df['best'])),
                    hovertemplate=_template_best())

    y_axis_label = f"{trials[0].objective.type.capitalize()} '{trials[0].objective.name}'"
    fig.update_layout(title=f"Regret for experiment '{experiment.name}'",
                      xaxis_title=f"Trials ordered by {order_by} time",
                      yaxis_title=y_axis_label)

    return fig


def _format_value(value):
    """
    Hyperparameter can have many types, sometimes they can even be lists.
    If one of the value is a float, it has to be compact.
    """
    if isinstance(value, str):
        return value

    try:
        return f"[{','.join(_format_value(x) for x in value)}]"

    except TypeError:
        if isinstance(value, float):
            return f"{value:.5G}"
        else:
            return value


def _format_hyperparameters(hyperparameters):
    result = ''

    for name, value in hyperparameters.items():
        x = f'<br>  {name[1:]}: {_format_value(value)}'
        result += x

    return result


def _template_trials(verbose_hover):
    template = '<b>ID: %{customdata[0]}</b><br>' \
        'value: %{y}<br>' \
        'time: %{customdata[1]|%Y-%m-%d %H:%M:%S}<br>'

    if verbose_hover:
        template += 'parameters: %{customdata[2]}'

    template += '<extra></extra>'

    return template


def _template_best():
    return '<b>Best ID: %{customdata[0]}</b><br>' \
        'value: %{customdata[1]}' \
        '<extra></extra>'



# -*- coding: utf-8 -*-
"""
:mod:`orion.core.worker` -- Coordination of the optimization procedure
======================================================================

.. module:: worker
   :platform: Unix
   :synopsis: Executes optimization steps and runs training experiment
      with parameter values suggested.

"""
import itertools
import logging

from orion.core.utils.exceptions import WaitingForTrials
from orion.core.utils.format_terminal import format_stats
from orion.core.worker.consumer import Consumer
from orion.core.worker.producer import Producer


log = logging.getLogger(__name__)


def reserve_trial(experiment, producer, _depth=1):
    """Reserve a new trial, or produce and reserve a trial if none are available."""
    trial = experiment.reserve_trial()

    if trial is None and not experiment.is_done:

        if _depth > 10:
            raise WaitingForTrials('No trials are available at the moment '
                                   'wait for current trials to finish')

        log.debug("#### Failed to pull a new trial from database.")

        log.debug("#### Fetch most recent completed trials and update algorithm.")
        producer.update()

        log.debug("#### Produce new trials.")
        producer.produce()

        return reserve_trial(experiment, producer, _depth=_depth + 1)

    return trial


COMPLETION_MESSAGE = """\
Hints
=====

Info
----

To get more information on the experiment, run the command

orion info --name {experiment.name} --version {experiment.version}

"""


NONCOMPLETED_MESSAGE = """\
Status
------

To get the status of the trials, run the command

orion status --name {experiment.name} --version {experiment.version}


For a detailed view with status of each trial listed, use the argument `--all`

orion status --name {experiment.name} --version {experiment.version} --all

"""


def workon(experiment, max_trials=None, max_broken=None, max_idle_time=None, heartbeat=None,
           user_script_config=None, interrupt_signal_code=None):
    """Try to find solution to the search problem defined in `experiment`."""
    producer = Producer(experiment, max_idle_time)
    consumer = Consumer(experiment, heartbeat, user_script_config, interrupt_signal_code)

    log.debug("#####  Init Experiment  #####")
    try:
        iterator = range(int(max_trials))
    except (OverflowError, TypeError):
        # When worker_trials is inf
        iterator = itertools.count()

    worker_broken_trials = 0
    for _ in iterator:
        log.debug("#### Poll for experiment termination.")
        if experiment.is_broken:
            print("#### Experiment has reached broken trials threshold, terminating.")
            break

        if experiment.is_done:
            print("#####  Search finished successfully  #####")
            break

        log.debug("#### Try to reserve a new trial to evaluate.")
        try:
            trial = reserve_trial(experiment, producer)
        except WaitingForTrials as ex:
            print("### Experiment failed to reserve new trials: {reason}, terminating. "
                  .format(reason=str(ex)))
            break

        if trial is not None:
            log.debug("#### Successfully reserved %s to evaluate. Consuming...", trial)
            success = consumer.consume(trial)
            if not success:
                worker_broken_trials += 1

        if worker_broken_trials >= max_broken:
            print("#### Worker has reached broken trials threshold, terminating.")
            print(worker_broken_trials, max_broken)
            break

    print('\n' + format_stats(experiment))

    print('\n' + COMPLETION_MESSAGE.format(experiment=experiment))
    if not experiment.is_done:
        print(NONCOMPLETED_MESSAGE.format(experiment=experiment))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Example usage and tests for :mod:`orion.core.io.resolve_config`."""

import hashlib
import logging
import os
import shutil
import socket

import git
import pytest

import orion.core
import orion.core.io.resolve_config as resolve_config


@pytest.fixture
def force_is_exe(monkeypatch):
    """Mock resolve_config to recognize any string as an executable script."""
    def is_exe(path):
        return True

    monkeypatch.setattr(resolve_config, "is_exe", is_exe)


@pytest.mark.usefixtures("empty_config")
def test_fetch_default_options():
    """Verify default options"""
    default_config = resolve_config.fetch_default_options()

    assert default_config['algorithms'] == 'random'
    assert default_config['database']['host'] == socket.gethostbyname(socket.gethostname())
    assert default_config['database']['name'] == 'orion'
    assert default_config['database']['type'] == 'MongoDB'
    assert default_config['database']['port'] == 27017

    assert default_config['max_trials'] == float('inf')
    assert default_config['name'] is None
    assert default_config['pool_size'] == 1


@pytest.mark.usefixtures("empty_config")
def test_socket_on_osx(monkeypatch):
    """Verify that default hostname is set properly on OSX"""
    default_config = resolve_config.fetch_default_options()
    assert default_config['database']['host'] == socket.gethostbyname(socket.gethostname())
    assert socket.gethostbyname(socket.gethostname()) != 'localhost'

    monkeypatch.setattr(socket, 'gethostname', lambda: 'wrong_name_on_osx')
    config = orion.core.build_config()
    resolve_config.config = config
    default_config = resolve_config.fetch_default_options()
    assert default_config['database']['host'] == 'localhost'


def test_fetch_env_vars():
    """Verify env vars are fetched properly"""
    env_vars_config = resolve_config.fetch_env_vars()
    assert env_vars_config == {'database': {}}

    db_name = "orion_test"

    os.environ['ORION_DB_NAME'] = db_name

    env_vars_config = resolve_config.fetch_env_vars()
    assert env_vars_config == {'database': {'name': 'orion_test'}}

    db_type = "MongoDB"
    os.environ['ORION_DB_TYPE'] = db_type

    env_vars_config = resolve_config.fetch_env_vars()
    assert env_vars_config == {'database': {'name': db_name, 'type': db_type}}


@pytest.mark.usefixtures("version_XYZ")
def test_fetch_metadata_orion_version():
    """Verify orion version"""
    metadata = resolve_config.fetch_metadata()
    assert metadata['orion_version'] == 'XYZ'


def test_fetch_metadata_executable_users_script(script_path):
    """Verify executable user script with absolute path"""
    metadata = resolve_config.fetch_metadata(user_args=[script_path])
    assert metadata['user_script'] == os.path.abspath(script_path)


def test_fetch_metadata_non_executable_users_script():
    """Verify executable user script keeps given path"""
    script_path = 'tests/functional/demo/orion_config.yaml'
    metadata = resolve_config.fetch_metadata(user_args=[script_path])
    assert metadata['user_script'] == script_path


def test_fetch_metadata_python_users_script(script_path):
    """Verify user script is correctly infered if called with python"""
    metadata = resolve_config.fetch_metadata(user_args=['python', script_path, 'some', 'args'])
    assert metadata['user_script'] == script_path


def test_fetch_metadata_not_existed_path():
    """Verfiy the raise of error when user_script path does not exist"""
    path = 'dummy/path'
    with pytest.raises(OSError) as exc_info:
        resolve_config.fetch_metadata(user_args=[path])
    assert "The path specified for the script does not exist" in str(exc_info.value)


@pytest.mark.usefixtures()
def test_fetch_metadata_user_args(script_path):
    """Verify user args"""
    user_args = [os.path.abspath(script_path)] + list(map(str, range(10)))
    metadata = resolve_config.fetch_metadata(user_args=user_args)
    assert metadata['user_script'] == user_args[0]
    assert metadata['user_args'] == user_args


@pytest.mark.usefixtures("with_user_tsirif")
def test_fetch_metadata_user_tsirif():
    """Verify user name"""
    metadata = resolve_config.fetch_metadata()
    assert metadata['user'] == "tsirif"


def test_fetch_metadata():
    """Verify no additional data is stored in metadata"""
    metadata = resolve_config.fetch_metadata()
    len(metadata) == 4


def test_fetch_config_from_cmdargs():
    """Verify fetch_config returns empty dict on no config file path"""
    cmdargs = {
        'name': 'test',
        'user': 'me',
        'version': 1,
        'config': None,
        'exp_max_trials': 'exp_max_trials',
        'worker_trials': 'worker_trials',
        'exp_max_broken': 'exp_max_broken',
        'working_dir': 'working_dir',
        'pool_size': 'pool_size',
        'max_trials': 'max_trials',
        'heartbeat': 'heartbeat',
        'worker_max_trials': 'worker_max_trials',
        'worker_max_broken': 'worker_max_broken',
        'max_idle_time': 'max_idle_time',
        'interrupt_signal_code': 'interrupt_signal_code',
        'user_script_config': 'user_script_config',
        'manual_resolution': 'manual_resolution',
        'non_monitored_arguments': 'non_monitored_arguments',
        'ignore_code_changes': 'ignore_code_changes',
        'auto_resolution': 'auto_resolution',
        'branch_from': 'branch_from',
        'algorithm_change': 'algorithm_change',
        'code_change_type': 'code_change_type',
        'cli_change_type': 'cli_change_type',
        'branch_to': 'branch_to',
        'config_change_type': 'config_change_type'}

    config = resolve_config.fetch_config_from_cmdargs(cmdargs)

    assert config.pop('config', None) is None

    exp_config = config.pop('experiment')
    assert exp_config.pop('name') == 'test'
    assert exp_config.pop('version') == 1
    assert exp_config.pop('user') == 'me'
    assert exp_config.pop('max_trials') == 'exp_max_trials'
    assert exp_config.pop('max_broken') == 'exp_max_broken'
    assert exp_config.pop('working_dir') == 'working_dir'
    assert exp_config.pop('pool_size') == 'pool_size'

    assert exp_config == {}

    worker_config = config.pop('worker')
    assert worker_config.pop('heartbeat') == 'heartbeat'
    assert worker_config.pop('max_trials') == 'worker_max_trials'
    assert worker_config.pop('max_broken') == 'worker_max_broken'
    assert worker_config.pop('max_idle_time') == 'max_idle_time'
    assert worker_config.pop('interrupt_signal_code') == 'interrupt_signal_code'
    assert worker_config.pop('user_script_config') == 'user_script_config'

    assert worker_config == {}

    evc_config = config.pop('evc')
    assert evc_config.pop('manual_resolution') == 'manual_resolution'
    assert evc_config.pop('non_monitored_arguments') == 'non_monitored_arguments'
    assert evc_config.pop('ignore_code_changes') == 'ignore_code_changes'
    assert evc_config.pop('auto_resolution') == 'auto_resolution'
    assert evc_config.pop('branch_from') == 'branch_from'
    assert evc_config.pop('algorithm_change') == 'algorithm_change'
    assert evc_config.pop('code_change_type') == 'code_change_type'
    assert evc_config.pop('cli_change_type') == 'cli_change_type'
    assert evc_config.pop('branch_to') == 'branch_to'
    assert evc_config.pop('config_change_type') == 'config_change_type'

    assert evc_config == {}

    assert config == {}


@pytest.mark.parametrize(
    "argument",
    ['config', 'user', 'user_args', 'name', 'version', 'branch_from', 'branch_to'])
def test_fetch_config_from_cmdargs_no_empty(argument):
    """Verify fetch_config returns only defined arguments."""
    config = resolve_config.fetch_config_from_cmdargs({})
    assert config == {}

    config = resolve_config.fetch_config_from_cmdargs({argument: None})
    assert config == {}

    config = resolve_config.fetch_config_from_cmdargs({argument: False})
    assert config == {}

    config = resolve_config.fetch_config_from_cmdargs({argument: 1})

    if argument in ['name', 'user', 'version']:
        assert config == {'experiment': {argument: 1}}
    elif argument in ['branch_from', 'branch_to']:
        assert config == {'evc': {argument: 1}}
    else:
        assert config == {argument: 1}


def test_fetch_config_no_hit():
    """Verify fetch_config returns empty dict on no config file path"""
    config = resolve_config.fetch_config({"config": ""})
    assert config == {}


def test_fetch_config(config_file):
    """Verify fetch_config returns valid dictionnary"""
    config = resolve_config.fetch_config({"config": config_file})

    assert config.pop('storage') == {
        'database': {
            'host': 'mongodb://user:pass@localhost',
            'name': 'orion_test',
            'type': 'mongodb'}}

    assert config.pop('experiment') == {
        'max_trials': 100,
        'name': 'voila_voici',
        'pool_size': 1,
        'algorithms': 'random',
        'strategy': 'NoParallelStrategy'}

    assert config == {}


def test_fetch_config_global_local_coherence(monkeypatch, config_file):
    """Verify fetch_config parses local config according to global config structure."""
    def mocked_config(file_object):
        return orion.core.config.to_dict()
    monkeypatch.setattr('yaml.safe_load', mocked_config)

    config = resolve_config.fetch_config({"config": config_file})

    # Test storage subconfig
    storage_config = config.pop('storage')
    database_config = storage_config.pop('database')
    assert storage_config.pop('type') == orion.core.config.storage.type
    assert storage_config == {}

    assert database_config.pop('host') == orion.core.config.storage.database.host
    assert database_config.pop('name') == orion.core.config.storage.database.name
    assert database_config.pop('port') == orion.core.config.storage.database.port
    assert database_config.pop('type') == orion.core.config.storage.database.type

    assert database_config == {}

    # Test experiment subconfig
    exp_config = config.pop('experiment')
    assert exp_config.pop('max_trials') == orion.core.config.experiment.max_trials
    assert exp_config.pop('max_broken') == orion.core.config.experiment.max_broken
    assert exp_config.pop('working_dir') == orion.core.config.experiment.working_dir
    assert exp_config.pop('pool_size') == orion.core.config.experiment.pool_size
    assert exp_config.pop('algorithms') == orion.core.config.experiment.algorithms
    assert exp_config.pop('strategy') == orion.core.config.experiment.strategy

    assert exp_config == {}

    # Test worker subconfig
    worker_config = config.pop('worker')
    assert worker_config.pop('heartbeat') == orion.core.config.worker.heartbeat
    assert worker_config.pop('max_trials') == orion.core.config.worker.max_trials
    assert worker_config.pop('max_broken') == orion.core.config.worker.max_broken
    assert worker_config.pop('max_idle_time') == orion.core.config.worker.max_idle_time
    assert (worker_config.pop('interrupt_signal_code') ==
            orion.core.config.worker.interrupt_signal_code)
    assert (worker_config.pop('user_script_config') ==
            orion.core.config.worker.user_script_config)

    assert worker_config == {}

    # Test evc subconfig
    evc_config = config.pop('evc')
    assert evc_config.pop('auto_resolution') == orion.core.config.evc.auto_resolution
    assert evc_config.pop('manual_resolution') == orion.core.config.evc.manual_resolution
    assert (evc_config.pop('non_monitored_arguments') ==
            orion.core.config.evc.non_monitored_arguments)
    assert evc_config.pop('ignore_code_changes') == orion.core.config.evc.ignore_code_changes
    assert evc_config.pop('algorithm_change') == orion.core.config.evc.algorithm_change
    assert evc_config.pop('code_change_type') == orion.core.config.evc.code_change_type
    assert evc_config.pop('cli_change_type') == orion.core.config.evc.cli_change_type
    assert evc_config.pop('config_change_type') == orion.core.config.evc.config_change_type

    assert evc_config == {}

    # Confirm that all fields were tested.
    assert config == {}


def test_fetch_config_dash(monkeypatch, config_file):
    """Verify fetch_config supports dash."""
    def mocked_config(file_object):
        return {'experiment': {'max-broken': 10, 'algorithms': {'dont-change': 'me'}}}
    monkeypatch.setattr('yaml.safe_load', mocked_config)

    config = resolve_config.fetch_config({"config": config_file})

    assert config == {'experiment': {'max_broken': 10, 'algorithms': {'dont-change': 'me'}}}


def test_fetch_config_underscore(monkeypatch, config_file):
    """Verify fetch_config supports underscore as well."""
    def mocked_config(file_object):
        return {'experiment': {'max_broken': 10, 'algorithms': {'dont-change': 'me'}}}
    monkeypatch.setattr('yaml.safe_load', mocked_config)

    config = resolve_config.fetch_config({"config": config_file})

    assert config == {'experiment': {'max_broken': 10, 'algorithms': {'dont-change': 'me'}}}


def test_fetch_config_deprecated_max_trials(monkeypatch, config_file):
    """Verify fetch_config will overwrite deprecated value if also properly defined."""
    def mocked_config(file_object):
        return {'experiment': {'max_trials': 10}, 'max_trials': 20}
    monkeypatch.setattr('yaml.safe_load', mocked_config)

    config = resolve_config.fetch_config({"config": config_file})

    assert config == {'experiment': {'max_trials': 10}}


def test_fetch_config_deprecate_tricky_names(monkeypatch, config_file):
    """Verify fetch_config assigns values properly for the similar names."""
    def mocked_config(file_object):
        return {
            'experiment': {
                'worker_trials': 'should_be_ignored'},
            'max_trials': 'exp_max_trials',
            'max_broken': 'exp_max_broken',
            'worker_trials': 'worker_max_trials',
            'name': 'exp_name'
        }
    monkeypatch.setattr('yaml.safe_load', mocked_config)

    config = resolve_config.fetch_config({"config": config_file})

    assert config == {
        'experiment': {
            'name': 'exp_name',
            'max_trials': 'exp_max_trials',
            'max_broken': 'exp_max_broken'},
        'worker': {'max_trials': 'worker_max_trials'}}


def test_merge_configs_update_two():
    """Ensure update on first level is fine"""
    a = {'a': 1, 'b': 2}
    b = {'a': 3}

    m = resolve_config.merge_configs(a, b)

    assert m == {'a': 3, 'b': 2}


def test_merge_configs_update_three():
    """Ensure two updates on first level is fine"""
    a = {'a': 1, 'b': 2}
    b = {'a': 3}
    c = {'b': 4}

    m = resolve_config.merge_configs(a, b, c)

    assert m == {'a': 3, 'b': 4}


def test_merge_configs_update_four():
    """Ensure three updates on first level is fine"""
    a = {'a': 1, 'b': 2}
    b = {'a': 3}
    c = {'b': 4}
    d = {'a': 5, 'b': 6}

    m = resolve_config.merge_configs(a, b, c, d)

    assert m == {'a': 5, 'b': 6}


def test_merge_configs_extend_two():
    """Ensure extension on first level is fine"""
    a = {'a': 1, 'b': 2}
    b = {'c': 3}

    m = resolve_config.merge_configs(a, b)

    assert m == {'a': 1, 'b': 2, 'c': 3}


def test_merge_configs_extend_three():
    """Ensure two extensions on first level is fine"""
    a = {'a': 1, 'b': 2}
    b = {'c': 3}
    c = {'d': 4}

    m = resolve_config.merge_configs(a, b, c)

    assert m == {'a': 1, 'b': 2, 'c': 3, 'd': 4}


def test_merge_configs_extend_four():
    """Ensure three extensions on first level is fine"""
    a = {'a': 1, 'b': 2}
    b = {'c': 3}
    c = {'d': 4}
    d = {'e': 5}

    m = resolve_config.merge_configs(a, b, c, d)

    assert m == {'a': 1, 'b': 2, 'c': 3, 'd': 4, 'e': 5}


def test_merge_configs_update_extend_two():
    """Ensure update and extension on first level is fine"""
    a = {'a': 1, 'b': 2}
    b = {'b': 3, 'c': 4}

    m = resolve_config.merge_configs(a, b)

    assert m == {'a': 1, 'b': 3, 'c': 4}


def test_merge_configs_update_extend_three():
    """Ensure two updates and extensions on first level is fine"""
    a = {'a': 1, 'b': 2}
    b = {'b': 3, 'c': 4}
    c = {'a': 5, 'd': 6}

    m = resolve_config.merge_configs(a, b, c)

    assert m == {'a': 5, 'b': 3, 'c': 4, 'd': 6}


def test_merge_configs_update_extend_four():
    """Ensure three updates and extensions on first level is fine"""
    a = {'a': 1, 'b': 2}
    b = {'b': 3, 'c': 4}
    c = {'a': 5, 'd': 6}
    d = {'d': 7, 'e': 8}

    m = resolve_config.merge_configs(a, b, c, d)

    assert m == {'a': 5, 'b': 3, 'c': 4, 'd': 7, 'e': 8}


def test_merge_sub_configs_update_two():
    """Ensure updating to second level is fine"""
    a = {'a': 1, 'b': 2}
    b = {'b': {'c': 3}}

    m = resolve_config.merge_configs(a, b)

    assert m == {'a': 1, 'b': {'c': 3}}


def test_merge_sub_configs_sub_update_two():
    """Ensure updating on second level is fine"""
    a = {'a': 1, 'b': {'c': 2}}
    b = {'b': {'c': 3}}

    m = resolve_config.merge_configs(a, b)

    assert m == {'a': 1, 'b': {'c': 3}}

    a = {'a': 1, 'b': {'c': 2, 'd': 3}}
    b = {'b': {'c': 4}}

    m = resolve_config.merge_configs(a, b)

    assert m == {'a': 1, 'b': {'c': 4, 'd': 3}}


def test_merge_sub_configs_sub_extend_two():
    """Ensure updating to third level from second level is fine"""
    a = {'a': 1, 'b': {'c': 2}}
    b = {'d': {'e': 3}}

    m = resolve_config.merge_configs(a, b)

    assert m == {'a': 1, 'b': {'c': 2}, 'd': {'e': 3}}

    a = {'a': 1, 'b': {'c': 2, 'd': 3}}
    b = {'b': {'e': {'f': 4}}}

    m = resolve_config.merge_configs(a, b)

    assert m == {'a': 1, 'b': {'c': 2, 'd': 3, 'e': {'f': 4}}}


def test_merge_sub_configs_update_three():
    """Ensure updating twice to third level from second level is fine"""
    a = {'a': 1, 'b': {'c': 2}}
    b = {'b': {'c': 3}}
    c = {'b': {'c': {'d': 4}}}

    m = resolve_config.merge_configs(a, b, c)

    assert m == {'a': 1, 'b': {'c': {'d': 4}}}

    a = {'a': 1, 'b': {'c': 2, 'd': 3}}
    b = {'b': {'c': 4}}
    c = {'b': {'c': {'e': 5}}}

    m = resolve_config.merge_configs(a, b, c)

    assert m == {'a': 1, 'b': {'c': {'e': 5}, 'd': 3}}


@pytest.fixture
def repo():
    """Create a dummy repo for the tests."""
    os.chdir('../')
    os.makedirs('dummy_orion')
    os.chdir('dummy_orion')
    repo = git.Repo.init('.')
    with open('README.md', 'w+') as f:
        f.write('dummy content')
    repo.git.add('README.md')
    repo.index.commit('initial commit')
    repo.create_head('master')
    repo.git.checkout('master')
    yield repo
    os.chdir('../')
    shutil.rmtree('dummy_orion')
    os.chdir('orion')


def test_infer_versioning_metadata_on_clean_repo(repo):
    """
    Test how `infer_versioning_metadata` fills its different fields
    when the user's repo is clean:
    `is_dirty`, `active_branch` and `diff_sha`.
    """
    vcs = resolve_config.infer_versioning_metadata('.git')
    assert not vcs['is_dirty']
    assert vcs['active_branch'] == 'master'
    # the diff should be empty so the diff_sha should be equal to the diff sha of an empty string
    assert vcs['diff_sha'] == hashlib.sha256(''.encode('utf-8')).hexdigest()


def test_infer_versioning_metadata_on_dirty_repo(repo):
    """
    Test how `infer_versioning_metadata` fills its different fields
    when the uers's repo is dirty:
    `is_dirty`, `HEAD_sha`, `active_branch` and `diff_sha`.
    """
    existing_metadata = {}
    existing_metadata['user_script'] = '.git'
    vcs = resolve_config.infer_versioning_metadata('.git')
    repo.create_head('feature')
    repo.git.checkout('feature')
    with open('README.md', 'w+') as f:
        f.write('dummy dummy content')
    vcs = resolve_config.infer_versioning_metadata('.git')
    assert vcs['is_dirty']
    assert vcs['active_branch'] == 'feature'
    assert vcs['diff_sha'] != hashlib.sha256(''.encode('utf-8')).hexdigest()
    repo.git.add('README.md')
    commit = repo.index.commit('Added dummy_file')
    vcs = resolve_config.infer_versioning_metadata('.git')
    assert not vcs['is_dirty']
    assert vcs['HEAD_sha'] == commit.hexsha
    assert vcs['diff_sha'] == hashlib.sha256(''.encode('utf-8')).hexdigest()


def test_fetch_user_repo_on_non_repo(caplog):
    """Test if `fetch_user_repo` logs a warning when user's script is not a git repo."""
    with caplog.at_level(logging.WARNING):
        resolve_config.fetch_user_repo('.')
    messages = [rec.message for rec in caplog.records]
    assert len(messages) == 1
    assert "Script {} is not in a git repository".format(os.getcwd()) in messages[0]


def test_infer_versioning_metadata_on_detached_head(repo):
    """Test in the case of a detached head."""
    with open('README.md', 'w+') as f:
        f.write('dummy contentt')
    repo.git.add('README.md')
    repo.index.commit('2nd commit')
    existing_metadata = {}
    existing_metadata['user_script'] = '.git'
    repo.head.reference = repo.commit('HEAD~1')
    assert repo.head.is_detached
    vcs = resolve_config.infer_versioning_metadata('.git')
    assert vcs['active_branch'] is None



# -*- coding: utf-8 -*-
"""
:mod:`orion.client.cli` -- Helper function for returning results from script
============================================================================

.. module:: cli
   :platform: Unix
   :synopsis: Provides functions for communicating with `orion.core`.

"""
import os
import sys

from orion.core import config


IS_ORION_ON = False
_HAS_REPORTED_RESULTS = False
RESULTS_FILENAME = os.getenv('ORION_RESULTS_PATH', None)
if RESULTS_FILENAME and os.path.isfile(RESULTS_FILENAME):
    import json
    IS_ORION_ON = True

if RESULTS_FILENAME and not IS_ORION_ON:
    raise RuntimeWarning("Results file path provided in environmental variable "
                         "does not correspond to an existing file.")


def interrupt_trial():
    """Send interrupt signal to Oríon worker"""
    sys.exit(config.worker.interrupt_signal_code)


def report_objective(objective, name='objective'):
    """Report only the objective at the end of execution

    To send more data (statistic, constraint, gradient), use ``report_results``.

    .. warning::

       To be called only once in order to report a final evaluation of a particular trial.

    .. warning::

       Oríon is only minimizing. Make sure to report a metric that you seek to minimize.

    .. note::

       In case that user's script is not running in a orion's context,
       this function will act as a Python `print` function.

    Parameters
    ----------
    objective: float
        Objective the return to Oríon for the current trial.
    name: str, optional
        Name of the objective. Default is 'objective'.

    """
    report_results([dict(name=name, type='objective', value=objective)])


def report_bad_trial(objective=1e10, name='objective', data=None):
    """Report a bad trial with large objective to Oríon.

    This is especially useful if some parameter values lead to exceptions such as out of memory.
    Reporting a large objective from such trials will push algorithms towards valid
    configurations.

    .. warning::

       To be called only once in order to report a final evaluation of a particular trial.

    .. warning::

       Oríon is only minimizing. Make sure to report a metric that you seek to minimize.

    .. note::

       In case that user's script is not running in a orion's context,
       this function will act as a Python `print` function.

    Parameters
    ----------
    objective: float
        Objective the return to Oríon for the current trial. The default objective is 1e10.
        This may not be valid for some metrics and this value should be overrided accordingly. In
        the case of error rates for instance, the value should be 1.0.
    name: str, optional
        Name of the objective. Default is 'objective'.
    data: list of dict, optional
        A list of dictionary representing the results in the form
        dict(name=result_name, type='statistic', value=0). The types supported are
        'contraint', 'gradient' and 'statistic'.

    """
    if data is None:
        data = []
    report_results([dict(name=name, type='objective', value=objective)] + data)


def report_results(data):
    """Facilitate the reporting of results for a user's script acting as a
    black-box computation.

    .. warning::

       To be called only once in order to report a final evaluation of a particular trial.

    .. warning::

       Oríon is only minimizing. Make sure to report a metric that you seek to minimize.

    .. note::

       In case that user's script is not running in a orion's context,
       this function will act as a Python `print` function.

    Parameters
    ----------
    data: list of dict
        A list of dictionary representing the results in the form
        dict(name=result_name, type='statistic', value=0). The types supported are
        'objective', 'contraint', 'gradient' and 'statistic'. The list should contain at least
        one 'objective', which is the metric the algorithm will be minimizing.

    """
    global _HAS_REPORTED_RESULTS  # pylint:disable=global-statement
    if _HAS_REPORTED_RESULTS:
        raise RuntimeWarning("Has already reported evaluation results once.")
    if IS_ORION_ON:
        with open(RESULTS_FILENAME, 'w') as results_file:
            json.dump(data, results_file)
    else:
        print(data)
    _HAS_REPORTED_RESULTS = True



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Collection of tests for :mod:`orion.core.evc.adapters`."""

import pytest

from orion.algo.space import Real
from orion.core.evc.adapters import (
    Adapter, AlgorithmChange, CodeChange, CompositeAdapter, DimensionAddition, DimensionDeletion,
    DimensionPriorChange, DimensionRenaming)
from orion.core.io.space_builder import DimensionBuilder
from orion.core.worker.trial import Trial


@pytest.fixture
def dummy_param():
    """Give dummy param integer param with value 1"""
    return Trial.Param(name='dummy', type='integer', value=1)


@pytest.fixture
def small_prior():
    """Give string format of small uniform distribution prior"""
    return "uniform(0, 10)"


@pytest.fixture
def large_prior():
    """Give string format of large uniform distribution prior"""
    return "uniform(0, 1000)"


@pytest.fixture
def normal_prior():
    """Give string format of normal distribution prior"""
    return "normal(0, 1)"


@pytest.fixture
def disjoint_prior():
    """Give string format of uniform distribution disjoint from small and large one"""
    return "uniform(-20, -10)"


@pytest.fixture
def integer_prior():
    """Give string format of uniform distribution casted to integers"""
    return "uniform(-20, -10, discrete=True)"


@pytest.fixture
def categorical_prior():
    """Give string format of categorical prior with floats, integers and strings"""
    return "choices([0.1, 1, 2, 'string'])"


@pytest.fixture
def multidim_prior():
    """Give string format of real distribution with multiple dimensions"""
    return "uniform(0, 10, shape=(2, 2))"


@pytest.fixture
def trials(small_prior, large_prior, normal_prior, disjoint_prior,
           integer_prior, categorical_prior, multidim_prior):
    """Trials with dimensions for all priors defined as fixtures"""
    N_TRIALS = 10

    priors = dict((name, prior) for (name, prior) in locals().items()
                  if isinstance(name, str) and name.endswith("_prior"))

    trials = []
    for _ in range(N_TRIALS):
        params = []
        for name, prior in priors.items():
            dimension = DimensionBuilder().build(name, prior)
            value = dimension.sample()[0]
            params.append(Trial.Param(name=name, type=dimension.type, value=value).to_dict())
        trials.append(Trial(params=params))

    return trials


class TestDimensionAdditionInit(object):
    """Test initialization of :class:`orion.core.evc.adapters.DimensionAddition`"""

    def test_dimension_addition_init_with_param(self, dummy_param):
        """Test initialization of :class:`orion.core.evc.adapters.DimensionAddition`
        with valid param
        """
        dimension_addition_adapter = DimensionAddition(dummy_param)

        assert dimension_addition_adapter.param is dummy_param

    def test_dimension_addition_init_with_bad_param(self):
        """Test initialization of :class:`orion.core.evc.adapters.DimensionAddition`
        with object which is not a param or a dictionary definition of it
        """
        with pytest.raises(TypeError) as exc_info:
            DimensionAddition('bad')

        assert "Invalid param argument type ('<class 'str'>')." in str(exc_info.value)

    def test_dimension_addition_init_with_dict(self, dummy_param):
        """Test initialization of :class:`orion.core.evc.adapters.DimensionAddition`
        with a dictionary definition of a param
        """
        DimensionAddition(dummy_param.to_dict())

    def test_dimension_addition_init_with_bad_dict(self):
        """Test initialization of :class:`orion.core.evc.adapters.DimensionAddition`
        with a dictionary which is a bad definition of a param
        """
        with pytest.raises(AttributeError) as exc_info:
            DimensionAddition({'bad': 'dict'})

        assert "'Param' object has no attribute 'bad'" in str(exc_info.value)


class TestDimensionDeletionInit(object):
    """Test initialization of :class:`orion.core.evc.adapters.DimensionDeletion`"""

    def test_dimension_deletion_init_with_param(self, dummy_param):
        """Test initialization of :class:`orion.core.evc.adapters.DimensionDeletion`
        with valid param
        """
        dimension_deletion_adapter = DimensionDeletion(dummy_param)

        assert dimension_deletion_adapter.param is dummy_param

    def test_dimension_deletion_init_with_bad_param(self):
        """Test initialization of :class:`orion.core.evc.adapters.DimensionDeletion`
        with object which is not a param or a dictionary definition of it
        """
        with pytest.raises(TypeError) as exc_info:
            DimensionDeletion('bad')

        assert "Invalid param argument type ('<class 'str'>')." in str(exc_info.value)

    def test_dimension_deletion_init_with_dict(self, dummy_param):
        """Test initialization of :class:`orion.core.evc.adapters.DimensionDeletion`
        with a dictionary definition of a param
        """
        DimensionDeletion(dummy_param.to_dict())

    def test_dimension_deletion_init_with_bad_dict(self):
        """Test initialization of :class:`orion.core.evc.adapters.DimensionDeletion`
        with a dictionary which is a bad definition of a param
        """
        with pytest.raises(AttributeError) as exc_info:
            print(DimensionDeletion({'bad': 'dict'}).param)

        assert "'Param' object has no attribute 'bad'" in str(exc_info.value)


class TestDimensionPriorChangeInit(object):
    """Test initialization of :class:`orion.core.evc.adapters.DimensionPriorChange`"""

    def test_dimension_prior_change_init_with_dimensions(self, large_prior, small_prior):
        """Test initialization of :class:`orion.core.evc.adapters.DimensionPriorChange`
        with valid string definitions of dimension prior
        """
        dimension_prior_change = DimensionPriorChange('dummy', large_prior, small_prior)

        assert dimension_prior_change.old_prior == large_prior
        assert dimension_prior_change.new_prior == small_prior
        assert isinstance(dimension_prior_change.old_dimension, Real)
        assert isinstance(dimension_prior_change.new_dimension, Real)

        assert dimension_prior_change.old_dimension.interval() == (0, 1000)
        assert dimension_prior_change.new_dimension.interval() == (0, 10)

    def test_dimension_prior_change_init_with_bad_dimensions(self):
        """Test initialization of :class:`orion.core.evc.adapters.DimensionPriorChange`
        with non valid string definitions of dimension prior
        """
        with pytest.raises(TypeError) as exc_info:
            DimensionPriorChange('dummy', 'bad', 'priors')

        assert "Parameter 'old': Please provide a valid form for prior" in str(exc_info.value)


class TestDimensionRenamingInit(object):
    """Test initialization of :class:`orion.core.evc.adapters.DimensionRenaming`"""

    def test_dimension_renaming_init(self):
        """Test initialization of :class:`orion.core.evc.adapters.DimensionRenaming`
        with valid names
        """
        dimension_renaming = DimensionRenaming('old_name', 'new_name')
        assert dimension_renaming.old_name == 'old_name'
        assert dimension_renaming.new_name == 'new_name'

    def test_dimension_renaming_init_bad_name(self):
        """Test initialization of :class:`orion.core.evc.adapters.DimensionRenaming`
        with invalid names which are not strings
        """
        with pytest.raises(TypeError) as exc_info:
            DimensionRenaming({'bad': 'name'}, None)

        assert "" in str(exc_info.value)


class TestAlgorithmChangeInit(object):
    """Test initialization of :class:`orion.core.evc.adapters.AlgorithmChange`"""

    def test_algorithm_change_init(self):
        """Test initialization of :class:`orion.core.evc.adapters.AlgorithmChange`"""
        AlgorithmChange()


class TestCodeChangeInit(object):
    """Test initialization of :class:`orion.core.evc.adapters.CodeChange`"""

    def test_code_change_init(self):
        """Test initialization of :class:`orion.core.evc.adapters.CodeChange`
        with valid change types
        """
        code_change_adapter = CodeChange(CodeChange.NOEFFECT)
        assert code_change_adapter.change_type == CodeChange.NOEFFECT

        code_change_adapter = CodeChange(CodeChange.BREAK)
        assert code_change_adapter.change_type == CodeChange.BREAK

    def test_code_change_init_bad_type(self):
        """Test initialization of :class:`orion.core.evc.adapters.CodeChange`
        with invalid change types
        """
        with pytest.raises(ValueError) as exc_info:
            CodeChange('bad type')

        assert "Invalid code change type 'bad type'" in str(exc_info.value)


class TestCompositeAdapterInit(object):
    """Test initialization of :class:`orion.core.evc.adapters.CompositeAdapter`"""

    def test_composite_adapter_init_emtpy(self):
        """Test initialization of :class:`orion.core.evc.adapters.CompositeAdapter`
        with no adapters
        """
        composite_adapter = CompositeAdapter()
        assert len(composite_adapter.adapters) == 0

    def test_composite_adapter_init_with_adapters(self, dummy_param):
        """Test initialization of :class:`orion.core.evc.adapters.CompositeAdapter`
        with valid adapters
        """
        dimension_addition = DimensionAddition(dummy_param)
        dimension_deletion = DimensionDeletion(dummy_param)
        composite_adapter = CompositeAdapter(dimension_addition, dimension_deletion)
        assert len(composite_adapter.adapters) == 2
        assert isinstance(composite_adapter.adapters[0], DimensionAddition)
        assert isinstance(composite_adapter.adapters[1], DimensionDeletion)

    def test_composite_adapter_init_with_bad_adapters(self):
        """Test initialization of :class:`orion.core.evc.adapters.CompositeAdapter`
        with invalid adapters
        """
        with pytest.raises(TypeError) as exc_info:
            CompositeAdapter('bad', 'adapters')

        assert ("Provided adapters must be adapter objects, not '<class 'str'>" in
                str(exc_info.value))


def test_adapter_creation(dummy_param):
    """Test initialization using :meth:`orion.core.evc.adapters.Adapter.build`"""
    adapter = Adapter.build([{
        'of_type': 'DimensionAddition',
        'param': dummy_param.to_dict()}])

    assert isinstance(adapter, CompositeAdapter)
    assert len(adapter.adapters) == 1
    assert isinstance(adapter.adapters[0], DimensionAddition)
    assert adapter.adapters[0].param.to_dict() == dummy_param.to_dict()


class TestDimensionAdditionForwardBackward(object):
    """Test :meth:`orion.core.evc.adapters.DimensionAddition.forward` and
    :meth:`orion.core.evc.adapters.DimensionAddition.backward`
    """

    def test_dimension_addition_forward(self, trials):
        """Test :meth:`orion.core.evc.adapters.DimensionAddition.forward`
        with valid param and trials
        """
        new_param = Trial.Param(name='second_normal_prior', type='integer', value=1)
        dimension_addition_adapter = DimensionAddition(new_param)

        adapted_trials = dimension_addition_adapter.forward(trials)

        assert adapted_trials[0]._params[-1] == new_param
        assert adapted_trials[4]._params[-1] == new_param
        assert adapted_trials[-1]._params[-1] == new_param

    def test_dimension_addition_forward_already_existing(self, trials):
        """Test :meth:`orion.core.evc.adapters.DimensionAddition.forward`
        with valid param and incompatible trials because param already exists
        """
        new_param = Trial.Param(name='normal_prior', type='integer', value=1)
        dimension_addition_adapter = DimensionAddition(new_param)

        with pytest.raises(RuntimeError) as exc_info:
            dimension_addition_adapter.forward(trials)
        assert "Provided trial does not have a compatible configuration" in str(exc_info.value)

    def test_dimension_addition_backward(self, dummy_param, trials):
        """Test :meth:`orion.core.evc.adapters.DimensionAddition.backward`
        with valid param and valid trials
        """
        new_param = Trial.Param(name='second_normal_prior', type='integer', value=1)
        dimension_addition_adapter = DimensionAddition(new_param)

        sampler = DimensionBuilder().build('random', 'uniform(10, 100, discrete=True)')
        for trial in trials:
            random_param = new_param.to_dict()
            random_param['value'] = sampler.sample()
            trial._params.append(Trial.Param(**random_param))

        adapted_trials = dimension_addition_adapter.backward(trials)
        assert len(adapted_trials) == 0

        trials[0]._params[-1].value = 1
        assert trials[0]._params[-1] == new_param

        adapted_trials = dimension_addition_adapter.backward(trials)
        assert len(adapted_trials) == 1

        trials[4]._params[-1].value = 1
        assert trials[4]._params[-1] == new_param

        adapted_trials = dimension_addition_adapter.backward(trials)
        assert len(adapted_trials) == 2

        trials[-1]._params[-1].value = 1
        assert trials[-1]._params[-1] == new_param

        adapted_trials = dimension_addition_adapter.backward(trials)
        assert len(adapted_trials) == 3

        assert new_param not in (adapted_trials[0]._params)
        assert new_param not in (adapted_trials[1]._params)
        assert new_param not in (adapted_trials[2]._params)

    def test_dimension_addition_backward_not_existing(self, trials):
        """Test :meth:`orion.core.evc.adapters.DimensionAddition.backward`
        with valid param and invalid trials because param does not exist
        """
        new_param = Trial.Param(name='second_normal_prior', type='integer', value=1)
        dimension_addition_adapter = DimensionAddition(new_param)

        with pytest.raises(RuntimeError) as exc_info:
            dimension_addition_adapter.backward(trials)
        assert "Provided trial does not have a compatible configuration" in str(exc_info.value)


class TestDimensionDeletionForwardBackward(object):
    """Test :meth:`orion.core.evc.adapters.DimensionDeletion.forward` and
    :meth:`orion.core.evc.adapters.DimensionDeletion.backward`
    """

    def test_dimension_deletion_forward(self, trials):
        """Test :meth:`orion.core.evc.adapters.DimensionDeletion.forward`
        with valid param and valid trials
        """
        new_param = Trial.Param(name='second_normal_prior', type='integer', value=1)
        dimension_deletion_adapter = DimensionDeletion(new_param)

        sampler = DimensionBuilder().build('random', 'uniform(10, 100, discrete=True)')
        for trial in trials:
            random_param = new_param.to_dict()
            random_param['value'] = sampler.sample()
            trial._params.append(Trial.Param(**random_param))

        adapted_trials = dimension_deletion_adapter.forward(trials)
        assert len(adapted_trials) == 0

        trials[0]._params[-1].value = 1
        assert trials[0]._params[-1] == new_param

        adapted_trials = dimension_deletion_adapter.forward(trials)
        assert len(adapted_trials) == 1

        trials[4]._params[-1].value = 1
        assert trials[4]._params[-1] == new_param

        adapted_trials = dimension_deletion_adapter.forward(trials)
        assert len(adapted_trials) == 2

        trials[-1]._params[-1].value = 1
        assert trials[-1]._params[-1] == new_param

        adapted_trials = dimension_deletion_adapter.forward(trials)
        assert len(adapted_trials) == 3

        assert new_param not in (adapted_trials[0]._params)
        assert new_param not in (adapted_trials[1]._params)
        assert new_param not in (adapted_trials[2]._params)

    def test_dimension_deletion_forward_not_existing(self, trials):
        """Test :meth:`orion.core.evc.adapters.DimensionDeletion.forward`
        with valid param and invalid trials because param does not exist
        """
        new_param = Trial.Param(name='second_normal_prior', type='integer', value=1)
        dimension_deletion_adapter = DimensionDeletion(new_param)

        with pytest.raises(RuntimeError) as exc_info:
            dimension_deletion_adapter.forward(trials)
        assert "Provided trial does not have a compatible configuration" in str(exc_info.value)

    def test_dimension_deletion_backward(self, dummy_param, trials):
        """Test :meth:`orion.core.evc.adapters.DimensionDeletion.backward`
        with valid param and valid trials
        """
        new_param = Trial.Param(name='second_normal_prior', type='integer', value=1)
        dimension_deletion_adapter = DimensionDeletion(new_param)

        adapted_trials = dimension_deletion_adapter.backward(trials)

        assert adapted_trials[0]._params[-1] == new_param
        assert adapted_trials[4]._params[-1] == new_param
        assert adapted_trials[-1]._params[-1] == new_param

    def test_dimension_deletion_backward_already_existing(self, trials):
        """Test :meth:`orion.core.evc.adapters.DimensionDeletion.backward`
        with valid param and invalid trials because param already exist
        """
        new_param = Trial.Param(name='normal_prior', type='integer', value=1)
        dimension_deletion_adapter = DimensionDeletion(new_param)

        with pytest.raises(RuntimeError) as exc_info:
            dimension_deletion_adapter.backward(trials)
        assert "Provided trial does not have a compatible configuration" in str(exc_info.value)


class TestDimensionPriorChangeForwardBackward(object):
    """Test :meth:`orion.core.evc.adapters.DimensionPriorChange.forward` and
    :meth:`orion.core.evc.adapters.DimensionPriorChange.backward`
    """

    def test_dimension_prior_change_forward(self, large_prior, small_prior, trials):
        """Test :meth:`orion.core.evc.adapters.DimensionPriorChange.forward`
        with compatible priors
        """
        dimension_prior_change_adapter = DimensionPriorChange(
            'small_prior', small_prior, large_prior)

        adapted_trials = dimension_prior_change_adapter.forward(trials)

        assert len(adapted_trials) == len(trials)

    def test_dimension_prior_change_forward_incompatible_dimensions(
            self, small_prior, disjoint_prior, trials):
        """Test :meth:`orion.core.evc.adapters.DimensionPriorChange.forward`
        with incompatible priors, such that all trials are filtered out
        """
        dimension_prior_change_adapter = DimensionPriorChange(
            'small_prior', small_prior, disjoint_prior)

        adapted_trials = dimension_prior_change_adapter.forward(trials)

        assert len(adapted_trials) == 0

    def test_dimension_prior_change_backward(self, large_prior, small_prior, trials):
        """Test :meth:`orion.core.evc.adapters.DimensionPriorChange.backward`
        with compatible priors
        """
        dimension_prior_change_adapter = DimensionPriorChange(
            'large_prior', large_prior, small_prior)

        adapted_trials = dimension_prior_change_adapter.backward(trials)

        assert len(adapted_trials) == len(trials)

    def test_dimension_prior_change_backward_incompatible_dimensions(
            self, disjoint_prior, small_prior, trials):
        """Test :meth:`orion.core.evc.adapters.DimensionPriorChange.backward`
        with incompatible priors, such that all trials are filtered out
        """
        dimension_prior_change_adapter = DimensionPriorChange(
            'small_prior', disjoint_prior, small_prior)

        adapted_trials = dimension_prior_change_adapter.backward(trials)

        assert len(adapted_trials) == 0


class TestDimensionRenamingForwardBackward(object):
    """Test :meth:`orion.core.evc.adapters.DimensionRenaming.forward` and
    :meth:`orion.core.evc.adapters.DimensionRenaming.backward`
    """

    def test_dimension_renaming_forward(self, trials):
        """Test :meth:`orion.core.evc.adapters.DimensionRenaming.forward`
        with valid names
        """
        old_name = 'small_prior'
        new_name = 'new_name'

        dimension_renaming_adapter = DimensionRenaming(old_name, new_name)

        adapted_trials = dimension_renaming_adapter.forward(trials)

        assert len(adapted_trials) == len(trials)

        assert new_name in [param.name for param in adapted_trials[0]._params]
        assert old_name not in [param.name for param in adapted_trials[0]._params]

        assert new_name in [param.name for param in adapted_trials[-1]._params]
        assert old_name not in [param.name for param in adapted_trials[-1]._params]

    def test_dimension_renaming_forward_incompatible(self, trials):
        """Test :meth:`orion.core.evc.adapters.DimensionRenaming.forward`
        with non existing old name in trials
        """
        old_name = 'bad name'
        new_name = 'small_prior'

        dimension_renaming_adapter = DimensionRenaming(old_name, new_name)

        with pytest.raises(RuntimeError) as exc_info:
            dimension_renaming_adapter.forward(trials)
        assert "Provided trial does not have a compatible configuration" in str(exc_info.value)

    def test_dimension_renaming_backward(self, trials):
        """Test :meth:`orion.core.evc.adapters.DimensionRenaming.backward`
        with valid names
        """
        old_name = 'old_name'
        new_name = 'small_prior'

        dimension_renaming_adapter = DimensionRenaming(old_name, new_name)

        adapted_trials = dimension_renaming_adapter.backward(trials)

        assert len(adapted_trials) == len(trials)

        assert old_name in [param.name for param in adapted_trials[0]._params]
        assert new_name not in [param.name for param in adapted_trials[0]._params]

        assert old_name in [param.name for param in adapted_trials[-1]._params]
        assert new_name not in [param.name for param in adapted_trials[-1]._params]

    def test_dimension_renaming_backward_incompatible(self, trials):
        """Test :meth:`orion.core.evc.adapters.DimensionRenaming.backward`
        with non existing new name in trials
        """
        old_name = 'small_prior'
        new_name = 'bad name'

        dimension_renaming_adapter = DimensionRenaming(old_name, new_name)

        with pytest.raises(RuntimeError) as exc_info:
            dimension_renaming_adapter.backward(trials)
        assert "Provided trial does not have a compatible configuration" in str(exc_info.value)


class TestAlgorithmChangeForwardBackward(object):
    """Test :meth:`orion.core.evc.adapters.AlgorithmChange.forward` and
    :meth:`orion.core.evc.adapters.AlgorithmChange.backward`
    """

    def test_algorithm_change_forward(self, trials):
        """Test :meth:`orion.core.evc.adapters.AlgorithmChange.forward`"""
        algorithm_change_adaptor = AlgorithmChange()

        adapted_trials = algorithm_change_adaptor.forward(trials)

        assert len(adapted_trials) == len(trials)

        assert adapted_trials[0] is trials[0]
        assert adapted_trials[-1] is trials[-1]

    def test_algorithm_change_backward(self, trials):
        """Test :meth:`orion.core.evc.adapters.AlgorithmChange.backward`"""
        algorithm_change_adaptor = AlgorithmChange()

        adapted_trials = algorithm_change_adaptor.backward(trials)

        assert len(adapted_trials) == len(trials)

        assert adapted_trials[0] is trials[0]
        assert adapted_trials[-1] is trials[-1]


class TestCodeChangeForwardBackward(object):
    """Test :meth:`orion.core.evc.adapters.CodeChange.forward` and
    :meth:`orion.core.evc.adapters.CodeChange.backward`
    """

    def test_code_change_forward_noeffect(self, trials):
        """Test :meth:`orion.core.evc.adapters.CodeChange.forward` with change type NOEFFECT"""
        code_change_adapter = CodeChange(CodeChange.NOEFFECT)

        adapted_trials = code_change_adapter.forward(trials)

        assert len(adapted_trials) == len(trials)

        assert adapted_trials[0] is trials[0]
        assert adapted_trials[-1] is trials[-1]

    def test_code_change_forward_unsure(self, trials):
        """Test :meth:`orion.core.evc.adapters.CodeChange.forward` with change type UNSURE"""
        code_change_adapter = CodeChange(CodeChange.UNSURE)

        adapted_trials = code_change_adapter.forward(trials)

        assert len(adapted_trials) == len(trials)

        assert adapted_trials[0] is trials[0]
        assert adapted_trials[-1] is trials[-1]

    def test_code_change_forward_break(self, trials):
        """Test :meth:`orion.core.evc.adapters.CodeChange.forward` with change type BREAK"""
        code_change_adapter = CodeChange(CodeChange.BREAK)

        adapted_trials = code_change_adapter.forward(trials)

        assert len(adapted_trials) == 0

    def test_code_change_backward_noeffect(self, trials):
        """Test :meth:`orion.core.evc.adapters.CodeChange.backward` with change type NOEFFECT"""
        code_change_adapter = CodeChange(CodeChange.NOEFFECT)

        adapted_trials = code_change_adapter.backward(trials)

        assert len(adapted_trials) == len(trials)

        assert adapted_trials[0] is trials[0]
        assert adapted_trials[-1] is trials[-1]

    def test_code_change_backward_unsure(self, trials):
        """Test :meth:`orion.core.evc.adapters.CodeChange.backward` with change type UNSURE"""
        code_change_adapter = CodeChange(CodeChange.UNSURE)

        adapted_trials = code_change_adapter.backward(trials)

        assert len(adapted_trials) == 0

    def test_code_change_backward_break(self, trials):
        """Test :meth:`orion.core.evc.adapters.CodeChange.backward` with change type BREAK"""
        code_change_adapter = CodeChange(CodeChange.BREAK)

        adapted_trials = code_change_adapter.backward(trials)

        assert len(adapted_trials) == 0


class TestCompositeAdapterForwardBackward(object):
    """Test :meth:`orion.core.evc.adapters.CompositeAdapter.forward` and
    :meth:`orion.core.evc.adapters.CompositeAdapter.backward`
    """

    def test_composite_adapter_forward_emtpy(self, trials):
        """Test :meth:`orion.core.evc.adapters.CompositeAdapter.forward` and
        :meth:`orion.core.evc.adapters.CompositeAdapter.backward` with no adapters
        """
        composite_adapter = CompositeAdapter()

        assert len(composite_adapter.forward(trials)) == len(trials)
        assert len(composite_adapter.backward(trials)) == len(trials)

    def test_composite_adapter_forward(self, dummy_param, trials):
        """Test :meth:`orion.core.evc.adapters.CompositeAdapter.forward` with two adapters"""
        new_param = Trial.Param(name='second_normal_prior', type='integer', value=1)

        dimension_addition_adapter = DimensionAddition(new_param)
        dimension_deletion_adapter = DimensionDeletion(new_param)

        composite_adapter = CompositeAdapter(dimension_addition_adapter)

        adapted_trials = composite_adapter.forward(trials)

        assert adapted_trials[0]._params[-1] == new_param
        assert adapted_trials[4]._params[-1] == new_param
        assert adapted_trials[-1]._params[-1] == new_param

        composite_adapter = CompositeAdapter(dimension_addition_adapter, dimension_deletion_adapter)

        adapted_trials = composite_adapter.forward(trials)

        assert len(adapted_trials) == len(trials)

        assert new_param not in (adapted_trials[0]._params)
        assert new_param not in (adapted_trials[4]._params)
        assert new_param not in (adapted_trials[-1]._params)

    def test_composite_adapter_backward(self, dummy_param, trials):
        """Test :meth:`orion.core.evc.adapters.CompositeAdapter.backward` with two adapters"""
        new_param = Trial.Param(name='second_normal_prior', type='integer', value=1)

        dimension_addition_adapter = DimensionAddition(new_param)
        dimension_deletion_adapter = DimensionDeletion(new_param)

        composite_adapter = CompositeAdapter(dimension_deletion_adapter)

        adapted_trials = composite_adapter.backward(trials)

        assert adapted_trials[0]._params[-1] == new_param
        assert adapted_trials[4]._params[-1] == new_param
        assert adapted_trials[-1]._params[-1] == new_param

        composite_adapter = CompositeAdapter(dimension_addition_adapter, dimension_deletion_adapter)

        adapted_trials = composite_adapter.backward(trials)

        assert len(adapted_trials) == len(trials)

        assert new_param not in (adapted_trials[0]._params)
        assert new_param not in (adapted_trials[4]._params)
        assert new_param not in (adapted_trials[-1]._params)


def test_dimension_addition_configuration(dummy_param):
    """Test :meth:`orion.core.evc.adapters.DimensionAddition.configuration`"""
    dimension_addition_adapter = DimensionAddition(dummy_param)

    configuration = dimension_addition_adapter.configuration[0]

    assert configuration['of_type'] == "dimensionaddition"
    assert configuration['param'] == dummy_param.to_dict()

    assert Adapter.build([configuration]).adapters[0].configuration[0] == configuration


def test_dimension_deletion_configuration(dummy_param):
    """Test :meth:`orion.core.evc.adapters.DimensionDeletion.configuration`"""
    dimension_deletion_adapter = DimensionDeletion(dummy_param)

    configuration = dimension_deletion_adapter.configuration[0]

    assert configuration['of_type'] == "dimensiondeletion"
    assert configuration['param'] == dummy_param.to_dict()

    assert Adapter.build([configuration]).adapters[0].configuration[0] == configuration


def test_dimension_prior_change_configuration(small_prior, large_prior):
    """Test :meth:`orion.core.evc.adapters.DimensionPriorChange.configuration`"""
    dimension_name = 'small_prior'
    dimension_prior_change_adapter = DimensionPriorChange(dimension_name, small_prior, large_prior)

    configuration = dimension_prior_change_adapter.configuration[0]

    assert configuration['of_type'] == "dimensionpriorchange"
    assert configuration['name'] == dimension_name
    assert configuration['old_prior'] == small_prior
    assert configuration['new_prior'] == large_prior

    assert Adapter.build([configuration]).adapters[0].configuration[0] == configuration


def test_dimension_renaming_configuration():
    """Test :meth:`orion.core.evc.adapters.DimensionRenaming.configuration`"""
    old_name = 'old_name'
    new_name = 'new_name'
    dimension_renaming_adapter = DimensionRenaming(old_name, new_name)

    configuration = dimension_renaming_adapter.configuration[0]

    assert configuration['of_type'] == "dimensionrenaming"
    assert configuration['old_name'] == old_name
    assert configuration['new_name'] == new_name

    assert Adapter.build([configuration]).adapters[0].configuration[0] == configuration


def test_algorithm_change_configuration():
    """Test :meth:`orion.core.evc.adapters.AlgorithmChange.configuration`"""
    algorithm_change_adaptor = AlgorithmChange()

    configuration = algorithm_change_adaptor.configuration[0]

    assert configuration['of_type'] == "algorithmchange"

    assert Adapter.build([configuration]).adapters[0].configuration[0] == configuration


def test_code_change_configuration():
    """Test :meth:`orion.core.evc.adapters.CodeChange.configuration`"""
    code_change_adaptor = CodeChange(CodeChange.UNSURE)

    configuration = code_change_adaptor.configuration[0]

    assert configuration['of_type'] == "codechange"
    assert configuration['change_type'] == CodeChange.UNSURE

    assert Adapter.build([configuration]).adapters[0].configuration[0] == configuration


def test_composite_configuration(dummy_param):
    """Test :meth:`orion.core.evc.adapters.CompositeAdapter.configuration`"""
    new_param = Trial.Param(name='second_normal_prior', type='integer', value=1)
    dimension_addition_adapter = DimensionAddition(dummy_param)
    dimension_deletion_adapter = DimensionDeletion(new_param)

    composite_adapter = CompositeAdapter(dimension_addition_adapter, dimension_deletion_adapter)

    configuration = composite_adapter.configuration

    assert configuration[0] == dimension_addition_adapter.configuration[0]
    assert configuration[1] == dimension_deletion_adapter.configuration[0]

    assert Adapter.build(configuration).adapters[0].configuration[0] == configuration[0]
    assert Adapter.build(configuration).adapters[1].configuration[0] == configuration[1]



# -*- coding: utf-8 -*-
"""
:mod:`orion.algo.mutate_functions` --
Different mutate functions: large-scale evolution of image classifiers
===========================================================================================

.. module:: mutate_functions
    :platform: Unix
    :synopsis: Implement evolution to exploit configurations with fixed resource efficiently

"""
import logging

import numpy as np

logger = logging.getLogger(__name__)


def default_mutate(search_space, old_value, **kwargs):
    """Get a default mutate function"""
    multiply_factor = kwargs.pop('multiply_factor', 3.0)
    add_factor = kwargs.pop('add_factor', 1)
    volatility = kwargs.pop('volatility', 0.001)
    if search_space.type == "real":
        lower_bound, upper_bound = search_space.interval()
        factors = (1.0 / multiply_factor + (multiply_factor - 1.0 / multiply_factor) *
                   np.random.random())
        if lower_bound <= old_value * factors <= upper_bound:
            new_value = old_value * factors
        elif lower_bound > old_value * factors:
            new_value = lower_bound + volatility * np.random.random()
        else:
            new_value = upper_bound - volatility * np.random.random()
    elif search_space.type == "integer":
        lower_bound, upper_bound = search_space.interval()
        factors = int(add_factor * (2 * np.random.randint(2) - 1))
        if lower_bound <= old_value + factors <= upper_bound:
            new_value = int(old_value) + factors
        elif lower_bound > old_value + factors:
            new_value = int(lower_bound)
        else:
            new_value = int(upper_bound)
    elif search_space.type == "categorical":
        sample_index = \
            np.where(np.random.multinomial(1,
                                           list(search_space.get_prior)) == 1)[0][0]
        new_value = int(search_space.categories[sample_index])
    else:
        new_value = old_value
    return new_value



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Perform a functional test of the info command."""
import orion.core.cli
import orion.core.io.resolve_config


def test_info_no_hit(clean_db, one_experiment, capsys):
    """Test info if no experiment with given name."""
    returncode = orion.core.cli.main(['info', '--name', 'i do not exist'])

    assert returncode == 1

    captured = capsys.readouterr().err

    assert captured.startswith('Error: No experiment with given name \'i do not exist\'')


def test_info_hit(clean_db, one_experiment, capsys):
    """Test info if existing experiment."""
    orion.core.cli.main(['info', '--name', 'test_single_exp'])

    captured = capsys.readouterr().out

    assert '--x~uniform(0,1)' in captured


def test_info_broken(clean_db, broken_refers, capsys):
    """Test info if experiment.refers is missing."""
    orion.core.cli.main(['info', '--name', 'test_single_exp'])

    captured = capsys.readouterr().out

    assert '--x~uniform(0,1)' in captured


def test_info_no_branching(clean_db, one_experiment_changed_vcs, capsys):
    """Test info if config file is different

    Version should not increase!
    """
    orion.core.cli.main(['info', '--name', 'test_single_exp'])

    captured = capsys.readouterr().out

    assert '\nversion: 1\n' in captured


def test_info_python_api(clean_db, with_experiment_using_python_api, capsys):
    """Test info if config built using python api"""
    orion.core.cli.main(['info', '--name', 'from-python-api'])

    captured = capsys.readouterr().out

    assert 'from-python-api' in captured
    assert 'Commandline' not in captured


def test_missing_conf_file(clean_db, with_experiment_missing_conf_file, capsys):
    """Test info can handle experiments when the user script config file is missing"""
    orion.core.cli.main(['info', '--name', 'test_single_exp'])

    captured = capsys.readouterr().out

    assert '--x~uniform(0,1)' in captured


def test_info_cmdline_api(clean_db, with_experiment_using_python_api, capsys):
    """Test info if config built using cmdline api"""
    orion.core.cli.main(['info', '--name', 'test_single_exp'])

    captured = capsys.readouterr().out

    assert 'test_single_exp' in captured
    assert 'Commandline' in captured


def test_no_args(capsys):
    """Try to run the command without any arguments"""
    returncode = orion.core.cli.main(["info"])
    assert returncode == 1

    captured = capsys.readouterr().err

    assert captured == 'Error: No name provided for the experiment.\n'



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Example usage and tests for :mod:`orion.core.io.space_builder`."""
import pytest
from scipy.stats import distributions as dists

from orion.algo.space import (Categorical, Fidelity, Integer, Real)
from orion.core.io.space_builder import (DimensionBuilder, SpaceBuilder)


@pytest.fixture(scope='module')
def dimbuilder():
    """Return a `DimensionBuilder` instance."""
    return DimensionBuilder()


@pytest.fixture(scope='module')
def spacebuilder():
    """Return a `SpaceBuilder` instance."""
    return SpaceBuilder()


class TestDimensionBuilder(object):
    """Ways of Dimensions builder."""

    def test_build_loguniform(self, dimbuilder):
        """Check that loguniform is built into reciprocal correctly."""
        dim = dimbuilder.build('yolo', 'loguniform(0.001, 10)')
        assert isinstance(dim, Real)
        assert dim.name == 'yolo'
        assert dim._prior_name == 'reciprocal'
        assert 3.3 in dim and 11.1 not in dim
        assert isinstance(dim.prior, dists.reciprocal_gen)

        dim = dimbuilder.build('yolo2', 'loguniform(1, 1000, discrete=True)')
        assert isinstance(dim, Integer)
        assert dim.name == 'yolo2'
        assert dim._prior_name == 'reciprocal'
        assert 3 in dim and 0 not in dim and 3.3 not in dim
        assert isinstance(dim.prior, dists.reciprocal_gen)

    def test_eval_nonono(self, dimbuilder):
        """Make malevolent/naive eval access more difficult. I think."""
        with pytest.raises(RuntimeError):
            dimbuilder.build('la', "__class__")

    def test_build_a_good_real(self, dimbuilder):
        """Check that non registered names are good, as long as they are in
        `scipy.stats.distributions`.
        """
        dim = dimbuilder.build('yolo2', 'alpha(0.9, low=0, high=10, shape=2)')
        assert isinstance(dim, Real)
        assert dim.name == 'yolo2'
        assert dim._prior_name == 'alpha'
        assert 3.3 not in dim
        assert (3.3, 11.1) not in dim
        assert (3.3, 6) in dim
        assert isinstance(dim.prior, dists.alpha_gen)

    def test_build_a_good_integer(self, dimbuilder):
        """Check that non registered names are good, as long as they are in
        `scipy.stats.distributions`.
        """
        dim = dimbuilder.build('yolo3', 'poisson(5)')
        assert isinstance(dim, Integer)
        assert dim.name == 'yolo3'
        assert dim._prior_name == 'poisson'
        assert isinstance(dim.prior, dists.poisson_gen)

    def test_build_a_good_real_discrete(self, dimbuilder):
        """Check that non registered names are good, as long as they are in
        `scipy.stats.distributions`.
        """
        dim = dimbuilder.build('yolo3', 'alpha(1.1, discrete=True)')
        assert isinstance(dim, Integer)
        assert dim.name == 'yolo3'
        assert dim._prior_name == 'alpha'
        assert isinstance(dim.prior, dists.alpha_gen)

    def test_build_a_good_fidelity(self, dimbuilder):
        """Check that a Fidelity dimension is correctly built."""
        dim = dimbuilder.build('epoch', 'fidelity(1, 16, 4)')
        assert isinstance(dim, Fidelity)
        assert dim.name == 'epoch'
        assert dim.low == 1
        assert dim.high == 16
        assert dim.base == 4
        assert dim._prior_name == 'None'
        assert dim.prior is None

    def test_build_fidelity_default_base(self, dimbuilder):
        """Check that a Fidelity dimension is correctly built with default base."""
        dim = dimbuilder.build('epoch', 'fidelity(1, 16)')
        assert isinstance(dim, Fidelity)
        assert dim.name == 'epoch'
        assert dim.low == 1
        assert dim.high == 16
        assert dim.base == 2
        assert dim._prior_name == 'None'
        assert dim.prior is None

    def test_build_fails_because_of_name(self, dimbuilder):
        """Build fails because distribution name is not supported..."""
        with pytest.raises(TypeError) as exc:
            dimbuilder.build('yolo3', 'lalala(1.1, discrete=True)')
        assert 'Parameter' in str(exc.value)
        assert 'supported' in str(exc.value)

    def test_build_fails_because_of_unexpected_args(self, dimbuilder):
        """Build fails because argument is not supported..."""
        with pytest.raises(TypeError) as exc:
            dimbuilder.build('yolo3', 'alpha(1.1, whatisthis=5, discrete=True)')
        assert 'Parameter' in str(exc.value)
        assert 'unexpected' in str(exc.value.__cause__)

    def test_build_fails_because_of_ValueError_on_run(self, dimbuilder):
        """Build fails because ValueError happens on init."""
        with pytest.raises(TypeError) as exc:
            dimbuilder.build('yolo2', 'alpha(0.9, low=5, high=6, shape=2)')
        assert 'Parameter' in str(exc.value)
        assert 'Improbable bounds' in str(exc.value.__cause__)

    def test_build_fails_because_of_ValueError_on_init(self, dimbuilder):
        """Build fails because ValueError happens on init."""
        with pytest.raises(TypeError) as exc:
            dimbuilder.build('yolo2', 'alpha(0.9, low=4, high=10, size=2)')
        assert 'Parameter' in str(exc.value)
        assert 'size' in str(exc.value.__cause__)

    def test_build_gaussian(self, dimbuilder):
        """Check that gaussian/normal/norm is built into reciprocal correctly."""
        dim = dimbuilder.build('yolo', 'gaussian(3, 5)')
        assert isinstance(dim, Real)
        assert dim.name == 'yolo'
        assert dim._prior_name == 'norm'
        assert isinstance(dim.prior, dists.norm_gen)

        dim = dimbuilder.build('yolo2', 'gaussian(1, 0.5, discrete=True)')
        assert isinstance(dim, Integer)
        assert dim.name == 'yolo2'
        assert dim._prior_name == 'norm'
        assert isinstance(dim.prior, dists.norm_gen)

    def test_build_normal(self, dimbuilder):
        """Check that gaussian/normal/norm is built into reciprocal correctly."""
        dim = dimbuilder.build('yolo', 'normal(0.001, 10)')
        assert isinstance(dim, Real)
        assert dim.name == 'yolo'
        assert dim._prior_name == 'norm'
        assert isinstance(dim.prior, dists.norm_gen)

        dim = dimbuilder.build('yolo2', 'normal(1, 0.5, discrete=True)')
        assert isinstance(dim, Integer)
        assert dim.name == 'yolo2'
        assert dim._prior_name == 'norm'
        assert isinstance(dim.prior, dists.norm_gen)

    def test_build_choices(self, dimbuilder):
        """Create correctly a `Categorical` dimension."""
        dim = dimbuilder.build('yolo', "choices('adfa', 1, 0.3, 'asaga', shape=4)")
        assert isinstance(dim, Categorical)
        assert dim.name == 'yolo'
        assert dim._prior_name == 'Distribution'
        assert isinstance(dim.prior, dists.rv_discrete)

        dim = dimbuilder.build('yolo', "choices(['adfa', 1])")
        assert isinstance(dim, Categorical)
        assert dim.name == 'yolo'
        assert dim._prior_name == 'Distribution'
        assert isinstance(dim.prior, dists.rv_discrete)

        dim = dimbuilder.build('yolo2', "choices({'adfa': 0.1, 3: 0.4, 5: 0.5})")
        assert isinstance(dim, Categorical)
        assert dim.name == 'yolo2'
        assert dim._prior_name == 'Distribution'
        assert isinstance(dim.prior, dists.rv_discrete)

        with pytest.raises(TypeError) as exc:
            dimbuilder.build('yolo2', "choices({'adfa': 0.1, 3: 0.4})")
        assert 'Parameter' in str(exc.value)
        assert 'sum' in str(exc.value.__cause__)

    def test_build_fails_because_empty_args(self, dimbuilder):
        """What happens if somebody 'forgets' stuff?"""
        with pytest.raises(TypeError) as exc:
            dimbuilder.build('yolo', "choices()")
        assert 'Parameter' in str(exc.value)
        assert 'categories' in str(exc.value)

        with pytest.raises(TypeError) as exc:
            dimbuilder.build('what', "alpha()")
        assert 'Parameter' in str(exc.value)
        assert 'positional' in str(exc.value.__cause__)

    def test_build_fails_because_troll(self, dimbuilder):
        """What happens if somebody does not fit regular expression expected?"""
        with pytest.raises(TypeError) as exc:
            dimbuilder.build('yolo', "lalalala")
        assert 'Parameter' in str(exc.value)
        assert 'form for prior' in str(exc.value)


class TestSpaceBuilder(object):
    """Check whether space definition from various input format is successful."""

    def test_configuration_rebuild(self, spacebuilder):
        """Test that configuration can be used to recreate a space."""
        prior = {'x': 'uniform(0, 10, discrete=True)',
                 'y': 'loguniform(1e-08, 1)',
                 'z': 'choices([\'voici\', \'voila\', 2])'}
        space = spacebuilder.build(prior)
        assert space.configuration == prior

    def test_subdict_dimensions(self, spacebuilder):
        """Test space can have hierarchical structure."""
        prior = {'a': {'x': 'uniform(0, 10, discrete=True)'},
                 'b': {'y': 'loguniform(1e-08, 1)',
                       'z': 'choices([\'voici\', \'voila\', 2])'}}
        space = spacebuilder.build(prior)
        assert len(space) == 3
        assert 'a.x' in space
        assert 'b.y' in space
        assert 'b.z' in space



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Script that will always interrupt trials."""
import argparse

from orion.client import interrupt_trial, report_bad_trial, report_objective  # noqa: F401


def no_report():
    """Do not report any result"""
    return


def execute():
    """Execute a simple pipeline as an example."""
    parser = argparse.ArgumentParser()
    parser.add_argument('fct', type=str)
    parser.add_argument('--name', type=str)
    parser.add_argument('--objective', type=str)
    parser.add_argument('--data', type=str)
    parser.add_argument('-x', type=float)

    inputs = parser.parse_args()

    kwargs = {}

    # Maybe its a float, maybe user made a mistake and report objective='name'
    try:
        inputs.objective = float(inputs.objective)
    except (ValueError, TypeError):
        pass

    for key, value in vars(inputs).items():
        if value is not None:
            kwargs[key] = value

    kwargs.pop('fct')
    kwargs.pop('x')

    if 'data' in kwargs:
        kwargs['data'] = [dict(name=kwargs['data'], type='constraint', value=1.0)]

    globals()[inputs.fct](**kwargs)


if __name__ == "__main__":
    execute()



# -*- coding: utf-8 -*-
"""
:mod:`orion.algo.hyperband` -- A Novel Bandit-Based Approach to Hyperparameter Optimization
===========================================================================================

.. module:: hyperband
    :platform: Unix
    :synopsis: Implement Hyperband to exploit configurations with fixed resource efficiently

"""
import copy
import hashlib
import logging

import numpy

from orion.algo.base import BaseAlgorithm
from orion.algo.space import Fidelity

logger = logging.getLogger(__name__)

REGISTRATION_ERROR = """
Bad fidelity level {fidelity}. Should be in {budgets}.
Params: {params}
"""

SPACE_ERROR = """
Hyperband cannot be used if space does not contain a fidelity dimension.
For more information on the configuration and usage of Hyperband, see
https://orion.readthedocs.io/en/develop/user/algorithms.html#hyperband
"""

BUDGET_ERROR = """
Cannot build budgets below max_resources;
(max: {}) - (min: {}) > (num_rungs: {})
"""


def compute_budgets(max_resources, reduction_factor):
    """Compute the budgets used for each execution of hyperband"""
    num_brackets = int(numpy.log(max_resources) / numpy.log(reduction_factor))
    budgets = []
    budgets_tab = {}  # just for display consideration
    for bracket_id in range(0, num_brackets + 1):
        bracket_budgets = []
        num_trials = int(numpy.ceil(int((num_brackets + 1) / (num_brackets - bracket_id + 1)) *
                                    (reduction_factor ** (num_brackets - bracket_id))))

        min_resources = max_resources / reduction_factor ** (num_brackets - bracket_id)
        for i in range(0, num_brackets - bracket_id + 1):
            n_i = int(num_trials / reduction_factor ** i)
            min_i = int(min_resources * reduction_factor ** i)
            bracket_budgets.append((n_i, min_i))

            if budgets_tab.get(i):
                budgets_tab[i].append((n_i, min_i))
            else:
                budgets_tab[i] = [(n_i, min_i)]

        budgets.append(bracket_budgets)

    display_budgets(budgets_tab, max_resources, reduction_factor)

    return budgets


def display_budgets(budgets_tab, max_resources, reduction_factor):
    """Display hyperband budget as a table in debug log"""
    num_brackets = len(budgets_tab[0])
    table_str = 'Display Budgets:\n'
    col_format_str = '{:<4}' + ' {:<12}' * num_brackets + '\n'
    col_title_list = ['i  '] + ['n_i  r_i'] * num_brackets
    col_sub_list = ['---'] + ['---------'] * num_brackets
    table_str += col_format_str.format(*col_sub_list)
    table_str += col_format_str.format(*col_title_list)
    table_str += col_format_str.format(*col_sub_list)

    total_trials = 0
    for key, values in budgets_tab.items():
        table_row = '{:<4} '.format(key)
        for value in values:
            n_i, r_i = value
            total_trials += n_i
            st = '{:<5} {:<7}'.format(n_i, r_i)
            table_row += st
        table_str += table_row + '\n'
    table_str += col_format_str.format(*col_sub_list)
    table_str += 'max resource={}, eta={}, trials number of one execution={}\n' \
        .format(max_resources, reduction_factor, total_trials)
    logger.debug(table_str)


class Hyperband(BaseAlgorithm):
    """Hyperband formulates hyperparameter optimization as a pure-exploration non-stochastic
    infinite-armed bandit problem where a predefined resource like iterations, data samples,
    or features is allocated to randomly sampled configurations.`

    For more information on the algorithm,
    see original paper at http://jmlr.org/papers/v18/16-558.html.

    Li, Lisha et al. "Hyperband: A Novel Bandit-Based Approach to Hyperparameter Optimization"
    Journal of Machine Learning Research, 18:1-52, 2018.

    Parameters
    ----------
    space: `orion.algo.space.Space`
        Optimisation space with priors for each dimension.
    seed: None, int or sequence of int
        Seed for the random number generator used to sample new trials.
        Default: ``None``
    repetitions: int
        Number of execution of Hyperband. Default is numpy.inf which means to
        run Hyperband until no new trials can be suggested.

    """

    def __init__(self, space, seed=None, repetitions=numpy.inf):
        self.brackets = []
        super(Hyperband, self).__init__(space, seed=seed, repetitions=repetitions)

        self.trial_info_wo_fidelity = {}  # Stores Point id (with no fidelity) -> Bracket

        self.points_in_suggest_call = {}

        try:
            fidelity_index = self.fidelity_index
        except IndexError:
            raise RuntimeError(SPACE_ERROR)

        fidelity_dim = space.values()[fidelity_index]

        self.min_resources = fidelity_dim.low
        self.max_resources = fidelity_dim.high
        self.reduction_factor = fidelity_dim.base

        # if self.reduction_factor < 2:
        #     raise AttributeError("Reduction factor for Hyperband needs to be at least 2.")

        self.repetitions = repetitions

        # Counter for how many times Hyperband been executed
        self.executed_times = 0

        if self.reduction_factor >= 2:
            self.budgets = compute_budgets(self.max_resources, self.reduction_factor)
            self.brackets = [
                Bracket(self, bracket_budgets, 1)
                for bracket_budgets in self.budgets
            ]
            self.seed_rng(seed)
        else:
            self.budgets = None
            self.brackets = None
            logger.warning("Reduction factor for Hyperband needs to be at least 2")

    def sample(self, num, bracket, buffer=10):
        """Sample new points from bracket"""
        sample_buffer = bracket.rungs[0]['n_trials'] * buffer
        samples = self.space.sample(sample_buffer, seed=bracket.seed)
        i = 0
        points = []
        while len(points) < num and i < sample_buffer:
            point = samples[i]

            _id_wo_fidelity = self.get_id(point)
            _bracket_observed = self.trial_info_wo_fidelity.get(_id_wo_fidelity)

            if _id_wo_fidelity not in self.points_in_suggest_call and \
                    (not _bracket_observed or
                     (_bracket_observed.repetition_id < bracket.repetition_id and
                      _bracket_observed.get_point_max_resource(point) <
                      bracket.rungs[0]['resources'])):
                # if no duplicated found or the duplicated found existing in previous hyperband
                # execution with less resource

                point = list(point)
                point[self.fidelity_index] = bracket.rungs[0]['resources']
                points.append(tuple(point))

                self.points_in_suggest_call[_id_wo_fidelity] = bracket

            i += 1

        return points

    def seed_rng(self, seed):
        """Seed the state of the random number generator.

        :param seed: Integer seed for the random number generator.
        """
        self.seed = seed
        for i, bracket in enumerate(self.brackets):
            bracket.seed_rng(self.executed_times + seed + i if seed is not None else None)
        self.rng = numpy.random.RandomState(seed)

    @property
    def state_dict(self):
        """Return a state dict that can be used to reset the state of the algorithm."""
        return {'rng_state': self.rng.get_state(), 'seed': self.seed,
                'executed_times': self.executed_times}

    def set_state(self, state_dict):
        """Reset the state of the algorithm based on the given state_dict

        :param state_dict: Dictionary representing state of an algorithm
        """
        self.seed_rng(state_dict['seed'])
        self.rng.set_state(state_dict['rng_state'])
        self.executed_times = state_dict['executed_times']

    def suggest(self, num=1):
        """Suggest a number of new sets of parameters.

        Sample new points until first rung is filled. Afterwards
        waits for all trials to be completed before promoting trials
        to the next rung.

        Parameters
        ----------
        num: int, optional
            Number of points to suggest. Defaults to 1.

        Returns
        -------
        list of points or None
            A list of lists representing points suggested by the algorithm. The algorithm may opt
            out if it cannot make a good suggestion at the moment (it may be waiting for other
            trials to complete), in which case it will return None.

        """
        samples = []
        # Temp persist of the points proposed inside this call
        self.points_in_suggest_call = {}
        for bracket in reversed(self.brackets):
            if not bracket.is_filled:
                samples += bracket.sample()

        if samples:
            return samples

        # All brackets are filled

        for bracket in reversed(self.brackets):
            if bracket.is_ready() and not bracket.is_done:
                samples += bracket.promote()

        if samples:
            return samples

        # Either all brackets are done or none are ready and algo needs to wait for some trials to
        # complete
        if len(self.trial_info_wo_fidelity) >= self.space.cardinality:
            logger.warning('The number of unique trials of bottom rungs exceeds the search space '
                           'cardinality %i, Hyperband algorithm exits.', self.space.cardinality)
        else:
            logger.warning('Hyeprband can not suggest new samples, exit.')

        return None

    def get_id(self, point, ignore_fidelity=True):
        """Compute a unique hash for a point based on params, without fidelity level by default."""
        _point = list(point)
        if ignore_fidelity:
            non_fidelity_dims = _point[0:self.fidelity_index]
            non_fidelity_dims.extend(_point[self.fidelity_index + 1:])
            _point = non_fidelity_dims

        return hashlib.md5(str(_point).encode('utf-8')).hexdigest()

    def _refresh_bracket(self):
        """Refresh bracket if one hyperband execution is done"""
        if all(bracket.is_done for bracket in self.brackets):
            self.executed_times = self.brackets[0].repetition_id
            logger.debug('Hyperband execution %i is done, required to execute %s times',
                         self.executed_times, str(self.repetitions))

            # Continue to the next execution if need
            if self.executed_times < self.repetitions:
                self.brackets = [
                    Bracket(self, bracket_budgets, self.executed_times + 1)
                    for bracket_budgets in self.budgets
                ]

    def _get_bracket(self, point):
        """Get the bracket of a point during observe"""
        fidelity = point[self.fidelity_index]
        _id_wo_fidelity = self.get_id(point)

        brackets = []
        for bracket in self.brackets:
            # If find same point in first rung of a bracket,
            # the point should register in this bracket
            if _id_wo_fidelity in bracket.rungs[0]['results']:
                brackets = [bracket]
                break

        if not brackets:
            # If the point show in current hyeprband execution the first time,
            # the bracket with same fidelity in the first rung should be used,
            # the assumption is that there is no duplicated points inside same hyperband execution.
            brackets = [bracket for bracket in self.brackets
                        if bracket.rungs[0]['resources'] == fidelity]

        if not brackets:
            raise ValueError("No bracket found for point {0} with fidelity {1}"
                             .format(_id_wo_fidelity, fidelity))

        if len(brackets) > 1:
            logger.warning("More than one bracket found for point %s, this should not happen",
                           str(point))

        bracket = brackets[0]

        return bracket

    def observe(self, points, results):
        """Observe evaluation `results` corresponding to list of `points` in
        space.

        A simple random sampler though does not take anything into account.
        """
        for point, result in zip(points, results):

            bracket = self._get_bracket(point)

            try:
                bracket.register(point, result['objective'])
            except IndexError:
                logger.warning('Point registered to wrong bracket. This is likely due '
                               'to a corrupted database, where trials of different fidelity '
                               'have a wrong timestamps.')
                continue

            _id_wo_fidelity = self.get_id(point)
            if _id_wo_fidelity not in self.trial_info_wo_fidelity or \
                    bracket.get_point_max_resource(point) > \
                    self.trial_info_wo_fidelity[_id_wo_fidelity].get_point_max_resource(point):
                # If the point show up the first time or it has the bigger resource value
                # compared with the same point in previous hyperband execution
                self.trial_info_wo_fidelity[_id_wo_fidelity] = bracket

            self._refresh_bracket()

    @property
    def is_done(self):
        """Return True, if all required execution been done."""
        if self.executed_times >= self.repetitions:
            return True
        return False

    @property
    def fidelity_index(self):
        """Compute the index of the point when fidelity is."""
        def _is_fidelity(dim):
            return (isinstance(dim, Fidelity) or
                    (hasattr(dim, 'original_dimension') and
                     isinstance(dim.original_dimension, Fidelity)))

        return [i for i, dim in enumerate(self.space.values()) if _is_fidelity(dim)][0]


class Bracket():
    """Bracket of rungs for the algorithm Hyperband.

    Parameters
    ----------
    hyperband: `Hyperband` algorithm
        The hyperband algorithm object which this bracket will be part of.
    budgets: list of tuple
        Each tuple gives the (n_trials, resource_budget) for the respective rung.
    repetition_id: int
        The id of hyperband execution this bracket belongs to

    """

    def __init__(self, hyperband, budgets, repetition_id):
        self.hyperband = hyperband
        self.rungs = [dict(resources=budget, n_trials=n_trials, results=dict())
                      for n_trials, budget in budgets]
        self.seed = None
        self.repetition_id = repetition_id

        logger.debug('Bracket budgets: %s', str(budgets))

        # points = hyperband.sample(compute_rung_sizes(reduction_factor, len(budgets))[0])
        # for point in points:
        #     self.register(point, None)

    @property
    def is_filled(self):
        """Return True if first rung with trials is filled"""
        return self.has_rung_filled(0)

    def get_point_max_resource(self, point):
        """Return the max resource value that has been tried for a point"""
        max_resource = 0
        _id_wo_fidelity = self.hyperband.get_id(point)
        for rung in self.rungs:
            if _id_wo_fidelity in rung['results']:
                max_resource = rung['resources']

        return max_resource

    def sample(self):
        """Sample a new trial with lowest fidelity"""
        should_have_n_trials = self.rungs[0]['n_trials']
        n_trials = len(self.rungs[0]['results'])
        return self.hyperband.sample(should_have_n_trials - n_trials, self)

    def register(self, point, objective):
        """Register a point in the corresponding rung"""
        fidelity = point[self.hyperband.fidelity_index]
        rungs = [rung['results'] for rung in self.rungs if rung['resources'] == fidelity]
        if not rungs:
            budgets = [rung['resources'] for rung in self.rungs]
            raise IndexError(REGISTRATION_ERROR.format(fidelity=fidelity, budgets=budgets,
                                                       params=point))

        rungs[0][self.hyperband.get_id(point)] = (objective, point)

    def get_candidates(self, rung_id):
        """Get a candidate for promotion"""
        if self.has_rung_filled(rung_id + 1):
            return []

        rung = self.rungs[rung_id]['results']
        next_rung = self.rungs[rung_id + 1]['results']

        rung = list(sorted((objective, point) for objective, point in rung.values()))

        should_have_n_trials = self.rungs[rung_id + 1]['n_trials']
        points = []
        i = 0
        while len(points) + len(next_rung) < should_have_n_trials:
            objective, point = rung[i]
            assert objective is not None
            _id = self.hyperband.get_id(point)
            if _id not in next_rung:
                points.append(point)
            i += 1

        return points

    @property
    def is_done(self):
        """Return True, if the last rung is filled."""
        return self.has_rung_filled(len(self.rungs) - 1)

    def has_rung_filled(self, rung_id):
        """Return True, if the rung[rung_id] is filled."""
        n_trials = len(self.rungs[rung_id]['results'])
        return n_trials >= self.rungs[rung_id]['n_trials']

    def is_ready(self, rung_id=None):
        """Return True, if the bracket is ready for next promote"""
        if rung_id is not None:
            return (self.has_rung_filled(rung_id) and
                    all(objective is not None
                        for objective, _ in self.rungs[rung_id]['results'].values()))

        is_ready = False
        for _rung_id in range(len(self.rungs)):
            if self.has_rung_filled(_rung_id):
                is_ready = self.is_ready(_rung_id)
            else:
                break

        return is_ready

    def promote(self):
        """Promote the first candidate that is found and return it

        The rungs are iterated over in reversed order, so that high rungs
        are prioritised for promotions. When a candidate is promoted, the loop is broken and
        the method returns the promoted point.

        .. note ::

            All trials are part of the rungs, for any state. Only completed trials
            are eligible for promotion, i.e., only completed trials can be part of top-k.
            Lookup for promotion in rung l + 1 contains trials of any status.

        """
        if self.is_done:
            return None

        for rung_id in range(len(self.rungs)):
            if self.has_rung_filled(rung_id + 1):
                continue

            if not self.is_ready(rung_id):
                return None

            points = []
            for candidate in self.get_candidates(rung_id):
                # pylint: disable=logging-format-interpolation
                logger.debug(
                    'Promoting {point} from rung {past_rung} with fidelity {past_fidelity} to '
                    'rung {new_rung} with fidelity {new_fidelity}'.format(
                        point=candidate, past_rung=rung_id,
                        past_fidelity=candidate[self.hyperband.fidelity_index],
                        new_rung=rung_id + 1, new_fidelity=self.rungs[rung_id + 1]['resources']))

                candidate = list(copy.deepcopy(candidate))
                candidate[self.hyperband.fidelity_index] = self.rungs[rung_id + 1]['resources']
                points.append(tuple(candidate))

            return points

        return None

    def seed_rng(self, seed):
        """Seed the state of the random number generator.

        :param seed: Integer seed for the random number generator.
        """
        self.seed = seed

    def __repr__(self):
        """Return representation of bracket with fidelity levels"""
        return 'Bracket(resource={}, repetition id={})' \
            .format([rung['resources'] for rung in self.rungs], self.repetition_id)



# -*- coding: utf-8 -*-
"""
:mod:`orion.analysis.regret` -- Provide tools to calculate regret
=================================================================

.. module:: orion.analysis.regret
   :platform: Unix
   :synopsis: Provide tools to calculate regret
"""
import pandas as pd


def regret(trials, names=('best', 'best_id')):
    """
    Calculates the regret for a collection of :class:`Trial`. The regret is calculated sequentially
    from the order of the collection.

    Parameters
    ----------
    trials: DataFrame or dict
        A dataframe of trials containing, at least, the columns 'objective' and 'id'. Or a dict
        equivalent.
    names:
        A tuple containing the names of the columns. Default is ('best', 'best-id').

    Returns
    -------
    A copy of the original dataframe with two new columns containing respectively the best value
    so far and its trial id.
    """
    if len(names) != 2:
        raise ValueError(f"`names` requires a tuple with 2 elements. {len(names)} provided.")

    df = pd.DataFrame(trials, copy=True)
    if df.empty:
        return df

    df[names[0]] = df['objective'].cummin()
    df[names[1]] = __get_best_ids(df, names[0])
    return df


def __get_best_ids(dataframe, best_name):
    """Links the cumulative best objectives with their respective ids"""
    best_id = None
    best_objective = None
    result = []

    for i, id in enumerate(dataframe.id):
        if dataframe.objective[i] == dataframe[best_name][i] \
                and dataframe.objective[i] != best_objective:
            best_id = id
            best_objective = dataframe.objective[i]
        result.append(best_id)

    return result
